import typing
from typing import List, Tuple, Iterable, Iterator, TypeVar, Any, Literal, Callable
from itertools import islice
import json

T = TypeVar("T")
U = TypeVar("U")


def ranked_argsort(lst: List[int]) -> List[int]:
    """
    Return a list of the same order in which the elements values correspond to their ranked values

    Arguments:
        lst: List on which to operate

    Returns:
        `List[int]`: List of value ranks of each element in order.
    """
    unique_values = sorted(set(lst))
    ranks = {v: i + 1 for i, v in enumerate(unique_values)}
    return [ranks[i] for i in lst]


# this makes any set serializable. This allows me to write to json
# consider: https://stackoverflow.com/questions/50916422/python-typeerror-object-of-type-int64-is-not-json-serializable
class SetEncoder(json.JSONEncoder):
    """
    Class to transform sets into lists to enable json serialization
    """

    def default(self, obj):
        if isinstance(obj, set):  # add more object types with elif isinstance etc.
            return list(obj)
        return super().default(obj)  # return parent class


# for running either single or multithread
class DummyPool:
    "Class to mimic a Threadpool. Used for single threaded runs."

    def map(self, function: Callable[[T], U], iterable: Iterable[T]) -> List[U]:
        return list(map(function, iterable))

    def imap(self, function: Callable[[T], U], iterable: Iterable[T]) -> Iterable[U]:
        return map(function, iterable)

    def imap_unordered(
        self, function: Callable[[T], U], iterable: Iterable[T]
    ) -> Iterable[U]:
        return map(function, iterable)

    def starmap(self, function, list_of_iterables):
        return list(function(*args) for args in list_of_iterables)

    def __enter__(self):
        return self

    def __exit__(self, exc_value, exc_type, traceback):
        return False


X = TypeVar("X", str, int, float, bool)  # Add more types if needed


def json_extract(obj: Any, key: X) -> List[X]:
    """
    Recursively fetch values from nested dictionary.

    Will also enter lists of dictionaries.

    Arguments:
        obj: Object instance from which to fetch values
        key: String, Int, float or bool key for which associated values will be fetched
    """
    arr: List[X] = []

    def extract(obj: Any, arr: List[X], key: X):
        """Recursively search for values of key in nested dictionary or list tree"""
        if isinstance(obj, dict):
            for k, v in obj.items():
                if k == key:
                    arr.append(v)
                elif isinstance(v, (dict, list)):
                    extract(v, arr, key)
        elif isinstance(obj, list):
            for item in obj:
                extract(item, arr, key)
        return arr

    values = extract(obj, arr, key)
    return values


@typing.overload
def chunks(iterable: Iterable[T], n: Literal[2]) -> Iterator[Tuple[T, T]]: ...


@typing.overload
def chunks(iterable: Iterable[T], n: Literal[3]) -> Iterator[Tuple[T, T, T]]: ...


def chunks(iterable: Iterable[T], n: int) -> Iterator[Tuple[T, ...]]:
    """
    Yield successive n-sized chunks from iterable.

    Arguments:
        iterable: Iterable to chunk
        n: Number of elements in each chunk
    """
    iterable = iter(iterable)
    while chunk := tuple(islice(iterable, n)):
        yield chunk


PROTEINOGENIC_AMINO_ACIDS = [
    "ALA",
    "ARG",
    "ASN",
    "ASP",
    "CYS",
    "GLN",
    "GLU",
    "GLY",
    "HIS",
    "ILE",
    "LEU",
    "LYS",
    "MET",
    "PHE",
    "PRO",
    "SER",
    "THR",
    "TRP",
    "TYR",
    "VAL",
]

# some common special residues in pdb structures
SPECIAL_AMINO_ACIDS = [
    "ASX",
    "GLX",
    "SEC",
    "PYL",
    "UNK",
    "MSE",
    "SEP",
    "TPO",
    "PTR",
    "HYP",
    "CME",
    "CSO",
    "CSD",
    "PCA",
    "MLY",
    "DAL",
    "DAR",
    "DSG",
    "ORN",
    "PTM",
]



"""Identitying catalytic residues in protein structures by matching catalytic templates
    derived from the Mechanism and Catalytic Site Atlas

See Also:
    *PAPER TITLE*
    Authors
    bioRxiv TAG
    `doi:goes_here <link_goes_here>`_

"""

__author__ = "Raymund Hackett"
__license__ = "MIT"
__version__ = "0.1.2"



from __future__ import annotations

import collections
import math
import warnings
import csv
import itertools
import io
import sys
import os
import json
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import List, Tuple, Dict, Optional, IO, ClassVar
from functools import cached_property
from dataclasses import dataclass, field
from pathlib import Path

import rich
import pyjess
from rich.progress import Progress, SpinnerColumn, TimeElapsedColumn

try:
    from importlib.resources import files as resource_files
except ImportError:
    from importlib_resources import files as resource_files  # type: ignore

from enzymm.template import Template, AnnotatedTemplate, Vec3, check_template
from enzymm.utils import chunks, ranked_argsort
from enzymm.utils import PROTEINOGENIC_AMINO_ACIDS, SPECIAL_AMINO_ACIDS

__all__ = [
    "LogisticRegressionModel",
    "Match",
    "Matcher",
]


@dataclass(frozen=True)
class LogisticRegressionModel:
    """
    Class for storing a Logistc Regression Model for predicting if a match is correct.

    f(Xn) = 1/(1+e^-(beta0 + beta1*x1 + beta2*x2 + ...))

    Trained on data for matches with a particular size at a particular pairwise distance

    Attributes:
        coeficients: `list` of `floats` beta coefficients
        intercept: `float` beta0 intercept
        threshold: `float` optimal threshold for this model
    """

    coefficents: List[float]
    intercept: float
    threshold: float


@dataclass
class Match:
    """
    Class for storing annotated PyJess hits.

    This class is a wrapper around `pyjess.Hit` and the original template object
    that was used for the query.

    Attributes:
        hit: `~pyjess.Hit` instance
        complete: `bool` If the query matched all other templates within the same cluster
        pairwise_distance: `float` Pairwise distance at which this match was found
        index: `int` internal index of this match. Default 0

    """

    hit: pyjess.Hit
    complete: bool = field(default=False)
    pairwise_distance: float = field(default=0)
    index: int = field(default=0)
    _logistic_regression_models: ClassVar[
        Dict[str, Dict[str, List[LogisticRegressionModel]]]
    ]

    def dumps(self, header: bool = False) -> str:
        """
        Dump `Match` to a string. Calls `Match.dump()`

        Arguments:
            header: if a header line should be dumped to the string too.
        """
        buffer = io.StringIO()
        self.dump(buffer, header=header)
        return (
            buffer.getvalue()
        )  # returns entire content temporary file object as a string

    def dump2pdb(
        self, file: IO[str], include_query: bool = False, transform: bool = False
    ):
        """
        Dump the 3D coordinates of the `Match` to a '.pdb' file.

        Arguments:
            file: file type object to write to
            include_query: `bool` If the full query molecule should be written too.
            transform: `bool` If the matched atoms should be written to the template reference frame.

        Note:
            The query is never transformed to the target reference frame.

        Note:
            By default, atoms are written in the coordinate reference frame of the query.
        """

        # TODO option to include template atoms too. esp. with --transform
        # this requires a dump method for pyjess.TemplateAtoms
        # if NOT transform, rotate the template atoms into the query reference frame.
        # if transform then in pymol once could fetch the template reference pdb
        # select the template atoms in the reference_pdb (selection called ref_site)
        # then call align ref_site, csa*** name of template
        # TODO once writing one match to a single file is done
        # with --transform and --include query, rotate the query to the template reference too
        # simply by using the roation matrix or by selecting the atoms in the query
        # and aligning to the matched residues object.

        def write_atom_line(atom: pyjess.Atom) -> str:
            one_char_elements = {
                "H",
                "B",
                "C",
                "N",
                "O",
                "F",
                "P",
                "S",
                "K",
                "V",
                "Y",
                "I",
                "W",
                "U",
            }
            if atom.element in one_char_elements:
                return f"ATOM  {atom.serial:>5}  {atom.name:<3s}{atom.altloc if atom.altloc is not None else '':<1}{atom.residue_name:<3}{atom.chain_id:>2}{atom.residue_number:>4}{atom.insertion_code:1s}   {atom.x:>8.3f}{atom.y:>8.3f}{atom.z:>8.3f}{atom.occupancy:>6.2f}{atom.temperature_factor:>6.2f}      {atom.segment:<4s}{atom.element:>2s} \n"
            else:
                return f"ATOM  {atom.serial:>5} {atom.name:<4s}{atom.altloc if atom.altloc is not None else '':<1}{atom.residue_name:<3}{atom.chain_id:>2}{atom.residue_number:>4}{atom.insertion_code:1s}   {atom.x:>8.3f}{atom.y:>8.3f}{atom.z:>8.3f}{atom.occupancy:>6.2f}{atom.temperature_factor:>6.2f}      {atom.segment:<4s}{atom.element:>2s} \n"

        if include_query:  # write the original query molecule too
            file.write(f"HEADER MOLECULE_ID {self.hit.molecule().id}\n")
            for atom in self.hit.molecule(transform=transform):
                file.write(write_atom_line(atom))
            file.write("END\n\n")

        file.write(
            f"HEADER {self.predicted_correct} MATCH {self.hit.molecule().id} {self.index}\n"
        )

        file.write(
            f'REMARK TEMPLATE_PDB {str(self.hit.template.pdb_id)}_{",".join(set(res.chain_id for res in self.hit.template.residues))}\n'
        )

        # alias for improved readability
        template = self.hit.template
        cluster = template.cluster

        if cluster:
            file.write(
                f"REMARK TEMPLATE CLUSTER {cluster.id}_{str(cluster.member)}_{str(cluster.size)}\n"
            )
        if template.represented_sites:
            file.write(f"REMARK TEMPLATE RESIDUES {template.template_id_string}\n")
        file.write(f"REMARK MOLECULE_ID {str(self.hit.molecule().id)}\n")
        file.write(f"REMARK MATCH INDEX {self.index}\n")

        if transform:
            file.write("REMARK TEMPLATE COORDINATE FRAME\n")
        else:
            file.write("REMARK QUERY COORDINATE FRAME\n")

        for atom in self.hit.atoms(
            transform=transform
        ):  # if transform == True then coordinates are transformed to the template reference frame
            file.write(write_atom_line(atom))
        file.write("END\n\n")

    def dump(
        self, file: IO[str], header: bool = False, predict_correctness: bool = True
    ):
        """
        Dump the information associated with a `Match` to a '.tsv' like line.

        Arguments:
            file: `file-like` object to write to
            header: `bool` If a header line should be written too
            predict_correctness: 'bool' Wether to predict if the match is correct or not

        Note:
            Coordinate information is not written.
        """
        writer = csv.writer(
            file, dialect="excel-tab", delimiter="\t", lineterminator="\n"
        )
        # aliases for improved readability
        template = self.hit.template
        cluster = self.hit.template.cluster

        if header:
            writer.writerow(
                [
                    "query_id",
                    "pairwise_distance",
                    "match_index",
                    "template_pdb_id",
                    "template_pdb_chains",
                    "template_cluster_id",
                    "template_cluster_member",
                    "template_cluster_size",
                    "template_effective_size",
                    "template_dimension",
                    "template_mcsa_id",
                    "template_uniprot_id",
                    "template_ec",
                    "template_cath",
                    "template_multimeric",
                    "query_multimeric",
                    "query_atom_count",
                    "query_residue_count",
                    "rmsd",
                    "log_evalue",
                    "orientation",
                    "preserved_order",
                    "completeness",
                    "predicted_correct",
                    "matched_residues",
                    "number_of_mutated_residues",
                    "number_of_side_chain_residues_(template,reference)",
                    "number_of_metal_ligands_(template,reference)",
                    "number_of_ptm_residues_(template, reference)",
                    "total_reference_residues",
                ]
            )

        content = [
            str(self.hit.molecule().id),
            str(self.pairwise_distance),
            str(self.index),
            str(template.pdb_id if template.pdb_id else ""),
            (",".join(set(res.chain_id for res in template.residues))),
            str(cluster.id if cluster else ""),
            str(self.hit.template.cluster.member if self.hit.template.cluster else ""),
            str(cluster.size if cluster else ""),
            str(template.effective_size),
            str(template.dimension),
            str(template.mcsa_id if template.mcsa_id else ""),
            str(template.uniprot_id if template.uniprot_id else ""),
            ",".join(template.ec if template.ec is not None else ""),
            ",".join(template.cath if template.cath else ""),
            str(template.multimeric),
            str(self.multimeric),
            str(self.query_atom_count),
            str(self.query_residue_count),
            str(round(self.hit.rmsd, 5)),
            str(round(self.hit.log_evalue, 5)),
            str(round(self.orientation, 5)),
            str(self.preserved_resid_order),
            str(self.complete),
            str(self.predicted_correct) if predict_correctness else "",
            (",".join("_".join(t) for t in self.matched_residues)),
        ]

        # check if the template was annotated with M-CSA information
        if isinstance(template, AnnotatedTemplate):
            content.extend(
                [
                    str(template.number_of_mutated_residues),
                    ",".join(str(i) for i in template.number_of_side_chain_residues),
                    ",".join(str(i) for i in template.number_of_metal_ligands),
                    ",".join(str(i) for i in template.number_of_ptm_residues),
                    str(template.total_reference_residues),
                ]
            )
        else:
            content.extend(["", "", "", "", "", ""])

        writer.writerow(content)

    def get_identifying_attributes(self) -> Tuple[int, int, int]:
        """
        `tuple` of (M-CSA id, cluster id and template dimension).`
        """
        # return the tuple (hit.template.m-csa, hit.template.cluster.id, hit.template.dimension)
        template = self.hit.template
        return (
            template.mcsa_id,
            template.cluster.id,
            template.dimension,
        )

    @property
    def predicted_correct(self) -> bool:
        """`bool`: If the match is predicted as correct based on logistic regression models."""
        if (
            str(self.hit.template.effective_size)
            not in self._logistic_regression_models
        ):  # Since no models for 5+ residue matches is provided, these are predicted as true
            return True
        else:
            try:
                predictions = []
                for model in self._logistic_regression_models[
                    str(self.hit.template.effective_size)
                ][str(self.pairwise_distance)]:
                    value = 1 / (
                        1
                        + math.e
                        ** -(
                            model.intercept
                            + model.coefficents[0] * self.hit.rmsd
                            + model.coefficents[1] * self.orientation
                        )
                    )
                    predictions.append(value >= model.threshold)

                # majority decision from all models
                return bool(
                    sum(predictions)
                    >= round(
                        (
                            len(
                                self._logistic_regression_models[
                                    str(self.hit.template.effective_size)
                                ][str(self.pairwise_distance)]
                            )
                            / 2
                        ),
                        0,
                    )
                )

            except KeyError as exc:
                raise KeyError(
                    f"Missing appropriate model parameters to predict correctness. Encountered either unexpected dictionary structure or no models for the pairwise distance {self.pairwise_distance} were provided"
                ) from exc
            except IndexError as exc:
                raise IndexError(
                    "Missing coefficients for both RMSD and Residue Orientation. Expecting models with 2 coeficients."
                ) from exc

    @cached_property
    def atom_triplets(self) -> List[Tuple[pyjess.Atom, pyjess.Atom, pyjess.Atom]]:
        """
        `list`: of `~pyjess.Atom` triplets belonging to the same matched query residue.
        """
        # list with matched residues
        # # Hit.atoms is a list of matched atoms with all info on residue numbers and residue chain ids and atom types, this should conserve order if Hit.atoms is a list!!!
        atom_triplets = []
        for atom_triplet in chunks(
            self.hit.atoms(transform=True), 3
        ):  # yield chunks of 3 atoms each, transform true because for angle calculation atoms need to be in template reference frame
            if len(atom_triplet) != 3:
                raise ValueError(
                    f"Failed to construct residues. Got only {len(atom_triplet)} ATOM lines"
                )
            # check if all three atoms belong to the same residue by adding a tuple of their residue defining properties to a set
            unique_residues = {
                (atom.residue_name, atom.chain_id, atom.residue_number)
                for atom in atom_triplet
            }
            if len(unique_residues) != 1:
                raise ValueError(
                    f"Mixed up atom triplets {unique_residues}. The atoms come from different residues!"
                )
            atom_triplets.append(atom_triplet)
        return atom_triplets

    @property
    def matched_residues(self) -> List[Tuple[str, str, str]]:
        """
        `list`:  with information on all matched query residues.
        Elements have are `tuple` (`~pyjess.Atom.residue_name`, `~pyjess.Atom.chain_id`, `~pyjessAtom.residue_number`)
        """
        return [
            (
                atom_triplet[0].residue_name,
                atom_triplet[0].chain_id,
                str(atom_triplet[0].residue_number),
            )
            for atom_triplet in self.atom_triplets
        ]

    @property
    def multimeric(self) -> bool:
        """
        `bool`: If the matched atoms in the query stem from multiple protein chains
        """
        # note that these are pyjess atom objects!
        return not all(
            atom.chain_id == self.hit.atoms()[0].chain_id for atom in self.hit.atoms()
        )

    @property
    def preserved_resid_order(self) -> bool:
        """
        `bool`: If the residues in the template and in the matched query structure have the same relative order.

        Note:
            This is a good filtering parameter but excludes hits on examples of convergent evolution or circular permutations

        Note:
            Will always return `False` if either template or query is multimeric
        """
        if self.hit.template.multimeric or self.multimeric:
            return False
        else:
            # Now extract relative atom order in hit
            return (
                ranked_argsort(
                    [
                        atom_triplet[0].residue_number
                        for atom_triplet in self.atom_triplets
                    ]
                )
                == self.hit.template.relative_order
            )

    @cached_property
    def match_vector_list(cls) -> List[Vec3]:
        """
        `list` of `Vec3`: of orientation vectors for each matched residue in the query
        """
        # !!! atom coordinates must be in template coordinate system!
        vector_list = []
        for residue_index, residue in enumerate(cls.hit.template.residues):
            first_atom_index, second_atom_index = residue.orientation_vector_indices
            if (
                second_atom_index == 9
            ):  # Calculate orientation vector going from middle_atom to mitpoint between side1 and side2
                middle_atom = cls.atom_triplets[residue_index][first_atom_index]
                side1, side2 = [
                    atom
                    for atom in cls.atom_triplets[residue_index]
                    if atom != middle_atom
                ]
                midpoint = (Vec3.from_xyz(side1) + Vec3.from_xyz(side2)) / 2
                vector_list.append(midpoint - Vec3.from_xyz(middle_atom))
            else:
                # Calculate orientation vector going from first_atom to second_atom_index
                first_atom = cls.atom_triplets[residue_index][first_atom_index]
                second_atom = cls.atom_triplets[residue_index][second_atom_index]
                vector_list.append(
                    Vec3.from_xyz(second_atom) - Vec3.from_xyz(first_atom)
                )
        return vector_list

    @property
    def template_vector_list(self) -> List[Vec3]:
        """
        `list` of `Vec3`: of orientation vectors for each residue in the template
        """
        return [res.orientation_vector for res in self.hit.template.residues]

    @property
    def orientation(self) -> float:  # average angle
        """
        `float`: The arithmetic mean of per-residue orientation angles
        for matched pairs of template and query residues
        """
        if len(self.template_vector_list) != len(self.match_vector_list):
            raise ValueError(
                "Vector lists for Template and matching Query structure had different lengths."
            )

        # now calculate the angle between the vector of the template and the query per residue
        angle_list = []
        for i in range(len(self.template_vector_list)):
            angle_list.append(
                self.template_vector_list[i].angle_to(self.match_vector_list[i])
            )
        return sum(angle_list) / len(angle_list)

    @property
    def query_atom_count(self) -> int:
        """
        `int`: The number of atoms in the query molecule
        """
        return len(self.hit.molecule())

    @property
    def query_residue_count(self) -> int:
        """
        `int`: The number of residues in the query molecule
        """
        all_residue_numbers = set()
        for atom in self.hit.molecule():
            if atom.residue_name in PROTEINOGENIC_AMINO_ACIDS + SPECIAL_AMINO_ACIDS:
                all_residue_numbers.add(atom.residue_number)
        return len(all_residue_numbers)


# Load the logistic regression models from the json into a dictionary
with resource_files(__package__).joinpath("data", "logistic_regression_models.json").open() as f:  # type: ignore
    logistic_regression_models: Dict[str, Dict[str, List[LogisticRegressionModel]]] = {}
    model_dict = json.load(f)

    for template_size, pairwise_dict in model_dict["match_size"].items():
        logistic_regression_models[template_size] = {}
        for pairwide_distance, model_list in pairwise_dict["pairwise_distance"].items():
            list_of_log_regression_models = []
            for param_dict in model_list["model_list"]:
                list_of_log_regression_models.append(
                    LogisticRegressionModel(
                        coefficents=param_dict["coef"],
                        intercept=param_dict["intercept"],
                        threshold=param_dict["threshold"],
                    )
                )
            logistic_regression_models[template_size][
                pairwide_distance
            ] = list_of_log_regression_models

    Match._logistic_regression_models = logistic_regression_models


def load_molecules(
    molecule_paths: List[Path], conservation_cutoff: float = 0, warn: bool = False
) -> List[pyjess.Molecule]:
    """Load query molecules"""
    molecules = []
    stem_counter: Dict[str, int] = collections.defaultdict(int)
    for molecule_path in molecule_paths:
        stem = Path(molecule_path).stem
        stem_counter[stem] += 1
        if stem_counter[stem] > 1:
            # In case the same stem occurs multiple times, create a unique ID using the stem and a running number starting from 2
            unique_id = f"{stem}_{stem_counter[stem]}"
        else:
            unique_id = stem

        molecule = pyjess.Molecule.load(
            str(molecule_path), id=unique_id
        )  # by default it will stop at ENDMDL
        if conservation_cutoff:
            molecule.conserved(conservation_cutoff)
            # molecule = molecule.conserved(args.conservation_cutoff)
            # conserved is a method called on a molecule object that returns a filtered molecule
            # atoms with a temperature-factor BELOW the conservation cutoff will be excluded
        if molecule:
            molecules.append(
                molecule
            )  # load a molecule and filter it by conservation_cutoff
        elif warn:
            warnings.warn(f"received an empty molecule from {molecule_path}")

    if not molecules and warn:
        warnings.warn("received no molecules from input")

    return molecules


class Matcher:
    """
    Class from which a query `~pyjess.Molecule` is matched to a `list` of `Template`.
    """

    _DEFAULT_JESS_PARAMS = {
        3: {"rmsd": 2, "distance": 0.9, "max_dynamic_distance": 0.9},
        4: {"rmsd": 2, "distance": 1.7, "max_dynamic_distance": 1.7},
        5: {"rmsd": 2, "distance": 2.0, "max_dynamic_distance": 2.0},
        6: {"rmsd": 2, "distance": 2.0, "max_dynamic_distance": 2.0},
        7: {"rmsd": 2, "distance": 2.0, "max_dynamic_distance": 2.0},
        8: {"rmsd": 2, "distance": 2.0, "max_dynamic_distance": 2.0},
    }

    def __init__(
        self,
        templates: List[Template],
        jess_params: Optional[Dict[int, Dict[str, float]]] = None,
        conservation_cutoff: int = 0,
        warn: bool = False,
        verbose: bool = False,
        skip_smaller_hits: bool = False,
        match_small_templates: bool = False,
        cpus: int = (
            len(os.sched_getaffinity(0))
            if sys.platform == "linux"
            else os.cpu_count() or 1
        ),
        filter_matches: bool = True,
        console: rich.console.Console | None = None,
    ):
        """
        Initialize a `Matcher` instance

        Arguments:
            templates: `list` of `Template` to match
            jess_params: `dict` Dictionary of PyJess parameters to apply. Will superseed defaults.
            conservation_cutoff: `float` Atoms below this cutoff will not be matched. Default 0.
            warn: `bool` If warnings about issues during matching should be printed. Default `False`
            verbose: `bool` If progress statements on matching should be printed. Default `False`
            skip_smaller_hits: `bool` Continue searching the query against smaller templates, after a match against any larger one was found. Default `False`
            match_small_templates: `bool` If matches for Templates with fewer than 3 side-chain residues should be reported. Default `False`
            cpus: `int` The number of cpus for multithreading. If 0 (default), use all. If <0 leave this number of threads free.
            filter_matches: `bool` If matches should be filtered by wether they are predicted to be correct. Default `True`

        Note:
            Default jess parameters depend on the size of the template::

                _DEFAULT_JESS_PARAMS = {
                    3: {"rmsd": 2, "distance": 0.9, "max_dynamic_distance": 0.9},
                    4: {"rmsd": 2, "distance": 1.7, "max_dynamic_distance": 1.7},
                    5: {"rmsd": 2, "distance": 2.0, "max_dynamic_distance": 2.0},
                    6: {"rmsd": 2, "distance": 2.0, "max_dynamic_distance": 2.0},
                    7: {"rmsd": 2, "distance": 2.0, "max_dynamic_distance": 2.0},
                    8: {"rmsd": 2, "distance": 2.0, "max_dynamic_distance": 2.0},
                }

        """

        self.templates = templates
        self.cpus = cpus
        self.conservation_cutoff = conservation_cutoff
        self.warn = warn
        self.verbose = verbose
        self.skip_smaller_hits = skip_smaller_hits
        self.match_small_templates = match_small_templates
        self.filter_matches = filter_matches
        self.jess_params = (
            self._DEFAULT_JESS_PARAMS if jess_params is None else jess_params
        )
        self.console = rich.console.Console(quiet=True) if console is None else console

        # # optional code to find duplicates
        # def find_duplicates(objects):
        #     d = collections.defaultdict(list)
        #     for obj in objects:
        #         d[obj].append(obj)

        #     for k,v in d.items():
        #         if len(v) > 1:
        #             for dup in v:
        #                 print(dup.__dict__)
        #             break

        unique_templates = set(self.templates)
        if len(unique_templates) < len(self.templates):
            # find_duplicates(self.templates)
            raise ValueError("Duplicate templates were found.")

        if self.cpus <= 0:
            os_cpu_count = (
                len(os.sched_getaffinity(0))
                if sys.platform == "linux"
                else os.cpu_count()
            )
            if os_cpu_count is not None:
                self.cpus = max(1, os_cpu_count + self.cpus)
            else:
                self.cpus = 1

        self.verbose_print(f"PyJess Version: {pyjess.__version__}")
        self.verbose_print(f"Running on {self.cpus} Thread(s)")
        self.verbose_print(f"Warnings are set to {self.warn}")
        self.verbose_print(
            f"Skip_smaller_hits search is set to {self.skip_smaller_hits}"
        )

        if self.conservation_cutoff:
            self.verbose_print(f"Conservation Cutoff set to {self.conservation_cutoff}")

        # check each template and if it passes add it to the dictionary of templates
        self.templates_by_effective_size: Dict[int, List[Template]] = (
            collections.defaultdict(list)
        )  # Dictionary of List of Template objects grouped by Template.effective_size as keys
        for template in templates:
            # skip smaller templates if match_small_templates is not set!
            if not self.match_small_templates and template.effective_size < 3:
                continue
            if check_template(
                template, warn=self.warn
            ):  # returns True if the Template passed all checks or if warn is set to False
                self.templates_by_effective_size[template.effective_size].append(
                    template
                )

        if self.verbose:
            template_number_dict: Dict[int, int] = {}
            for size, template_list in self.templates_by_effective_size.items():
                template_number_dict[size] = len(template_list)
            print(
                f"Templates by effective size: {collections.OrderedDict(sorted(template_number_dict.items()))}"
            )

        self.template_effective_sizes = list(self.templates_by_effective_size.keys())
        self.template_effective_sizes.sort(
            reverse=True
        )  # get a list of template_sizes in decending order

        # print a warning if match_small_templates was set.
        if self.warn and self.match_small_templates:
            smaller_sizes = [i for i in self.template_effective_sizes if i < 3]
            if smaller_sizes:
                small_templates = []
                for i in smaller_sizes:
                    small_templates.extend(self.templates_by_effective_size[i])

                warnings.warn(
                    f"{len(small_templates)} Templates with an effective size smaller than 3 defined sidechain residues were supplied.\nFor small templates Jess parameters for templates of 3 residues will be used."
                )

                self.verbose_print(
                    "The templates with the following ids are too small:"
                )
                self.verbose_print([st.id for st in small_templates])

    def verbose_print(self, *args):
        """
        Make a print statement only in verbose mode
        """
        if self.verbose:
            print(*args)

    def _get_jess_parameters(self, template_size: int) -> Tuple[float, float, float]:
        if template_size < 3:
            parameter_size = 3
        elif template_size > 8:
            parameter_size = 8
        else:
            parameter_size = template_size

        rmsd = self.jess_params[parameter_size]["rmsd"]
        distance = self.jess_params[parameter_size]["distance"]
        max_dynamic_distance = self.jess_params[parameter_size]["max_dynamic_distance"]

        return rmsd, distance, max_dynamic_distance

    @staticmethod
    def _check_completeness(matches: List[Match]) -> List[Match]:
        # only after all templates of a certain size have been scanned could we compute the complete tag
        # This requries cluster and mcsa.id to be set! Otherwise I assume there is no cluster and therefore the match is complete by default!
        groupable_matches = []
        lone_matches = []

        for match in matches:
            if (
                match.hit.template.mcsa_id is not None
                and match.hit.template.cluster is not None
            ):
                groupable_matches.append(match)
            else:
                lone_matches.append(match)

        grouped_matches = [
            list(g)
            for _, g in itertools.groupby(
                sorted(groupable_matches, key=Match.get_identifying_attributes),
                Match.get_identifying_attributes,
            )
        ]

        for cluster_matches in grouped_matches:
            # For each query check if all Templates assigned to the same cluster targeted that structure
            #
            # TODO report statistics on this: This percentage of queries had a complete active site as reported by the complete tag
            # Filter this by template clusters with >1 member of course or report seperately by the number of clustermembers
            # or say like: This template cluster was always complete while this template cluster was only complete X times out of Y Queries matched to one member
            #
            # check if all the cluster members up to and including cluster_size are present in the group,
            indexed_possible_cluster_members = list(range(cluster_matches[0].hit.template.cluster.size))  # type: ignore
            possible_cluster_members = [x + 1 for x in indexed_possible_cluster_members]

            found_cluster_members = [match.hit.template.cluster.member for match in cluster_matches]  # type: ignore
            found_cluster_members.sort()

            if found_cluster_members == possible_cluster_members:
                for match in cluster_matches:
                    match.complete = True

        for match in lone_matches:
            match.complete = True

        return matches

    @staticmethod
    def _run_jess(
        molecule: pyjess.Molecule,
        templates: List[Template],
        rmsd_threshold: float = 2.0,
        distance_cutoff: float = 1.5,
        max_dynamic_distance: float = 1.5,
        max_candidates: int = 10000,
    ) -> List[Match]:
        """`list` of `Match`: Match the `list` of `Template` to one `~pyjess.Molecule`"""

        # killswitch is controlled by max_candidates. Internal default is currently 1000
        # killswitch serves to limit the iterations in cases where the template would be too general,
        # and the program would run in an almost endless loop

        jess = pyjess.Jess(
            templates
        )  # Create a Jess instance and use it to query a molecule (a PDB structure) against the stored templates:
        query = jess.query(
            molecule=molecule,
            rmsd_threshold=rmsd_threshold,
            distance_cutoff=distance_cutoff,
            max_dynamic_distance=max_dynamic_distance,
            max_candidates=max_candidates,
            best_match=True,
            ignore_chain=True,
        )  # query is pyjess.Query object which is an iterator over pyjess.Hits

        # ignore_chain=True disables checks for chain relationship
        # i.e. if two atoms are on a different chain in the template
        # they must also be on different chains in the target

        # best_match=True reports only the single best match between template and target
        # For this to make sense consider that:
        # A template is not encoded as coordinates, rather as a set of constraints.
        # For example, it would not contain the exact positions of THR and ASN atoms,
        # but instructions like
        # "Cα of ASN should be X angstrom away from the Cα of THR plus the allowed distance."

        # Multiple solutions = Matches to a template, satisfying all constraints may therefore exist
        # Jess produces matches to templates by looking for any combination of atoms,
        # residue_types, elements etc. and ANY positions which satisfy the constraints in the template

        # thus the solutions that Jess finds are NOT identical to the template at all
        # rather they are all possible solutions to the set constraints.
        # Solutions may completely differ from the template geometry
        # or atom composition if allowed by the set constraints.
        # by setting best_match=True we turn on filtering by rmsd to return only the best match
        # for every molecule template pair. Currently best_match is not exposed to the user.
        # This should be the only use case (I think)

        # Thus we hope to return the one solution to the constraints
        # which most closely resembles the original template - this is not guaranteed of course

        matches: List[Match] = []
        for hit in query:  # hit is pyjess.Hit
            matches.append(Match(hit=hit, pairwise_distance=distance_cutoff))

        return matches

    def _single_query_run(
        self,
        molecule: pyjess.Molecule,
        templates: List[Template],
        rmsd_threshold: float,
        distance_cutoff: float,
        max_dynamic_distance: float,
        max_candidates: int = 10000,
    ) -> List[Match]:

        matches = self._run_jess(
            molecule=molecule,
            templates=templates,
            rmsd_threshold=rmsd_threshold,
            distance_cutoff=distance_cutoff,
            max_dynamic_distance=max_dynamic_distance,
            max_candidates=max_candidates,
        )
        self._check_completeness(matches)

        return matches

    def _filter_molecule_matches(
        self,
        all_matches: List[Match],
    ):

        # keep only matches predicted as correct
        if self.filter_matches:
            filtered_matches = []
            for match in all_matches:
                if match.predicted_correct:
                    filtered_matches.append(match)
            return filtered_matches

        # return unchanged
        else:
            return all_matches

    def _worker(
        self,
        mol: pyjess.Molecule,
        template_size: int,
    ):
        templates = self.templates_by_effective_size[template_size]
        rmsd, distance, max_dynamic_distance = self._get_jess_parameters(template_size)
        results = self._single_query_run(
            mol,
            templates=templates,
            rmsd_threshold=rmsd,
            distance_cutoff=distance,
            max_dynamic_distance=max_dynamic_distance,
        )

        return mol, self._filter_molecule_matches(results)

    def run(
        self,
        molecules: List[pyjess.Molecule],
    ) -> Dict[pyjess.Molecule, List[Match]]:
        """
        Run the matcher against a `list` of query `~pyjess.Molecule` to search.

        Arguments:
            molecules: `list` of `~pyjess.Molecule` to search

        Returns:
            `dict` of `~pyjess.Molecule` --> `list` of `Match`: Dictionary of query molecules as keys and all found matches as values.
        """

        processed_molecules: Dict[pyjess.Molecule, List[Match]] = (
            collections.defaultdict(list)
        )

        # state: a dict of molecule: template sizes which still need to be searched
        # Each molecule has a copy of the template size list
        remaining_sizes = {
            mol: self.template_effective_sizes.copy() for mol in molecules
        }

        with ThreadPoolExecutor(max_workers=self.cpus) as pool, Progress(
            SpinnerColumn(),
            *Progress.get_default_columns(),
            TimeElapsedColumn(),
            "Structures {task.completed}/{task.total}",
            console=self.console,
        ) as progress:

            task_id = progress.add_task(
                description="[green]Searching structures ...", total=len(molecules)
            )

            futures = {}
            # Seed jobs with the largest template size per molecule
            for mol, sizes in remaining_sizes.items():
                if sizes:
                    futures[pool.submit(self._worker, mol, sizes.pop(0))] = mol
                else:
                    progress.advance(task_id)

            while futures:
                for fut in as_completed(futures, timeout=None):
                    mol = futures.pop(fut)

                    # # if there is an error in the worker function, id like to see it
                    # results = None
                    # try:
                    #     mol, results = fut.result()
                    # except Exception as e:
                    #     raise e

                    # but for now i trust the worker function 100%
                    mol, results = fut.result()

                    if results:  # match found --> stop further searches with molecule
                        processed_molecules[mol].extend(results)
                        if self.skip_smaller_hits:
                            progress.advance(task_id)
                            remaining_sizes.pop(
                                mol, None
                            )  # remove to stop further jobs
                            continue

                    # no match --> schedule next smaller template for this molecule
                    sizes = remaining_sizes[mol]
                    if sizes:
                        next_size = sizes.pop(0)
                        futures[pool.submit(self._worker, mol, next_size)] = mol
                    else:
                        # Exhausted all sizes with no match
                        progress.advance(task_id)
                        remaining_sizes.pop(mol, None)

        return processed_molecules

    def run_single(self, molecule: pyjess.Molecule) -> List[Match]:
        """
        Run the matcher against a single query `~pyjess.Molecule`.

        Argument:
            molecule: `~pyjess.Molecule` to search

        Returns;
            `list`: of `Match` found for the query `~pyjess.Molecule`
        """
        return self.run([molecule])[molecule]



from __future__ import annotations

import os
import sys
import contextlib
import glob
import warnings
import io
import math
import re
import json
from pathlib import Path
from typing import (
    Any,
    Sequence,
    List,
    Tuple,
    Dict,
    Optional,
    Union,
    TextIO,
    Iterator,
    Iterable,
    ContextManager,
    ClassVar,
)
from functools import cached_property
from dataclasses import dataclass
from multiprocessing.pool import ThreadPool

import pyjess

try:
    from importlib.resources import files as resource_files
except ImportError:
    from importlib_resources import files as resource_files  # type: ignore

from enzymm.utils import chunks, ranked_argsort, DummyPool
from enzymm.mcsa_info import load_mcsa_catalytic_residue_homologs_info
from enzymm.mcsa_info import ReferenceCatalyticResidue, NonReferenceCatalyticResidue


@dataclass(frozen=True)
class Vec3:
    """
    Class for storing 3D vectors in XYZ.

    Attributes:
        x : `float` X-Coordinate value
        y : `float` Y-Coordinate value
        z : `float` Z-Coordinate value

    """

    x: float
    y: float
    z: float

    def __post_init__(self):
        if math.isnan(self.x) or math.isnan(self.y) or math.isnan(self.z):
            raise ValueError(
                "Cannot create a Vec3 with NaN values. Likely the Jess superposition failed."
            )

    @classmethod
    def from_xyz(cls, item: Any) -> Vec3:
        """
        Create a `Vec3` instance from an object with x,y,z attributes.

        Args:
            item `any` : Any object with x, y, z attributes

        Returns:
            `Vec3` instance
        """
        return Vec3(item.x, item.y, item.z)

    @property
    def norm(self) -> float:
        """
        `float`: The vector norm (root of sum of squares)
        """
        return math.sqrt(self.x**2 + self.y**2 + self.z**2)

    def normalize(self) -> Vec3:
        """
        `Vec3`: The vector devided by its norm
        """
        norm = self.norm
        if norm == 0:  # zero vector
            return Vec3.from_xyz(self)
        return self / norm

    def __matmul__(self, other: Vec3) -> float:
        """
        Overloads the @ operator to perform dot product between two `Vec3` vectors.

        Args:
            other: `Vec3` instance

        Returns:
            `Vec3`
        """
        if isinstance(other, Vec3):
            return self.x * other.x + self.y * other.y + self.z * other.z
        raise TypeError(f"Expected Vec3, got {type(other).__name__}")

    def __add__(self, other: Union[int, float, Vec3]) -> Vec3:
        """
        Overloads the + operator to add either an `int`, `float` or other `Vec3` to a `Vec3`.

        Args:
            other: `int` | `float` | `Vec3` argument to add

        Returns:
            `Vec3`
        """
        if isinstance(other, Vec3):
            return Vec3(self.x + other.x, self.y + other.y, self.z + other.z)
        elif isinstance(other, (int, float)):
            return Vec3(self.x + other, self.y + other, self.z + other)
        raise TypeError(f"Expected int, float or Vec3, got {type(other).__name__}")

    def __truediv__(self, other: Union[int, float, Vec3]) -> Vec3:
        """
        Overloads the / operator to divide a `Vec3` object by either an `int`, `float` or other `Vec3`.

        Args:
            other: `int` | `float` | `Vec3` argument to divide by

        Returns:
            `Vec3`
        """
        if isinstance(other, Vec3):
            return Vec3(self.x / other.x, self.y / other.y, self.z / other.z)
        elif isinstance(other, (int, float)):
            return Vec3(self.x / other, self.y / other, self.z / other)
        raise TypeError(f"Expected int, float or Vec3, got {type(other).__name__}")

    def __sub__(self, other: Union[int, float, Vec3]) -> Vec3:
        """
        Overloads the - operator to subtract either an `int`, `float` or other `Vec3` from a `Vec3`.

        Args:
            other: `int` | `float` | `Vec3` argument to subtract

        Returns:
            `Vec3`
        """
        if isinstance(other, Vec3):
            return Vec3(self.x - other.x, self.y - other.y, self.z - other.z)
        elif isinstance(other, (int, float)):
            return Vec3(self.x - other, self.y - other, self.z - other)
        raise TypeError(f"Expected int, float or Vec3, got {type(other).__name__}")

    # from https://stackoverflow.com/questions/2827393/angles-between-two-n-dimensional-vectors-in-python
    def angle_to(self, other: Vec3) -> float:
        """
        Returns the angle in radians between two `Vec3` vectors.

        Args:
            other: `Vec3` to which to calculate the angle

        Returns:
            `float`: angle in radians

        """
        dot_product = self.normalize() @ other.normalize()
        if -1 <= dot_product <= 1:
            return math.acos(dot_product)
        else:
            # due to numerical errors two identical vectors may have a dot_product not exactly 1
            if math.isclose(dot_product, 1, rel_tol=1e-5):
                return 0
            # same but with opposite vectors
            elif math.isclose(dot_product, -1, rel_tol=1e-5):
                return math.pi
            else:
                raise ValueError(
                    f"ArcCos is not defined outside [-1,1]. self.vec is {[self.x, self.y, self.z]}, other vec is {[other.x, other.y, other.z]}"
                )


@dataclass(frozen=True, init=False)
class Residue:
    """
    Class for storing template residues (defined as 3 atoms) with relevant information.
    """

    _atoms: Tuple[pyjess.TemplateAtom, pyjess.TemplateAtom, pyjess.TemplateAtom]
    _vec: Vec3
    _indices: Tuple[int, int]

    def __init__(
        self,
        atoms: Tuple[pyjess.TemplateAtom, pyjess.TemplateAtom, pyjess.TemplateAtom],
    ):
        """
        Initilaize a `Residue` instance from a triplet of `~pyjess.TemplateAtom` objects.

        Args:
            atoms: `tuple` of 3 `~pyjess.TemplateAtom` instances

        Returns:
            `Residue`
        """

        vec, indices = self.calc_residue_orientation(atoms)
        object.__setattr__(self, "_atoms", atoms)
        object.__setattr__(self, "_vec", vec)
        object.__setattr__(self, "_indices", indices)

    @staticmethod
    def calc_residue_orientation(
        atoms: Tuple[pyjess.TemplateAtom, pyjess.TemplateAtom, pyjess.TemplateAtom]
    ) -> Tuple[Vec3, Tuple[int, int]]:
        """
        Method to calculate the residue orientation depending on the residue type.

        Args:
            atoms: `tuple` of 3 `~pyjess.TemplateAtom`

        Note:
            For symmetric atom triplets, the angle is calculated from the central atom to the midpoint between the two identical atom types.
            For non-symetric atom triplets, the angle is calculated between two atoms following the axis of polarization.

        Returns:
            `tuple`:  of `Residue.orientation_vector` and `Residue.orientation_vector_indices`
        """

        # dictionary in which the vectors from start to finish are defined for each aminoacid type
        # the orientation vector is calculated differently for different aminoacid types
        vector_atom_type_dict = {
            "GLY": ("C", "O"),
            "ALA": ("CA", "CB"),
            "VAL": ("CA", "CB"),
            "LEU": ("CA", "CB"),
            "ILE": ("CA", "CB"),
            "MET": ("CG", "SD"),
            "PHE": ("CZ", "mid"),
            "TYR": ("CZ", "OH"),
            "TRP": ("CZ2", "NE1"),
            "CYS": ("CB", "SG"),
            "PRO": ("C", "O"),
            "SER": ("CB", "OG"),
            "THR": ("CB", "OG1"),
            "ASN": ("CG", "OD1"),
            "GLN": ("CD", "OE1"),
            "LYS": ("CE", "NZ"),
            "ARG": ("CZ", "mid"),
            "HIS": ("CG", "ND1"),
            "ASP": ("CG", "mid"),
            "GLU": ("CD", "mid"),
            "PTM": ("CA", "CB"),
            "ANY": ("C", "O"),
        }
        try:
            vectup = vector_atom_type_dict[atoms[0].residue_names[0]]
        except KeyError as exc:
            raise KeyError(
                f"Residue orientation is not defined for the residue type {atoms[0].residue_names[0]}"
            ) from exc

        # In residues with two identical atoms, the vector is calculated from the middle atom to the mid point between the identical pair
        if vectup[1] == "mid":
            try:
                middle_index, middle_atom = next(
                    (index, atom)
                    for index, atom in enumerate(atoms)
                    if atom.atom_names[0] == vectup[0]
                )
                side1, side2 = [atom for atom in atoms if atom != middle_atom]
                midpoint = (Vec3.from_xyz(side1) + Vec3.from_xyz(side2)) / 2
                return midpoint - Vec3.from_xyz(middle_atom), (middle_index, 9)
            except StopIteration:
                raise ValueError(
                    f"Failed to find middle atom for amino-acid {atoms[0].residue_names[0]!r}"
                ) from None

        else:  # from first atom to second atom
            try:
                first_atom_index, first_atom = next(
                    (index, atom)
                    for index, atom in enumerate(atoms)
                    if atom.atom_names[0] == vectup[0]
                )
            except StopIteration:
                raise ValueError(
                    f"Failed to find first atom for amino-acid {atoms[0].residue_names[0]!r}"
                ) from None
            try:
                second_atom_index, second_atom = next(
                    (index, atom)
                    for index, atom in enumerate(atoms)
                    if atom.atom_names[0] == vectup[1]
                )
            except StopIteration:
                raise ValueError(
                    f"Failed to find second atom for amino-acid {atoms[0].residue_names[0]!r}"
                ) from None
            return Vec3.from_xyz(second_atom) - Vec3.from_xyz(first_atom), (
                first_atom_index,
                second_atom_index,
            )

    @classmethod
    def construct_residues_from_atoms(
        cls, atoms: Iterable[pyjess.TemplateAtom]
    ) -> List[Residue]:

        residues = []
        for atom_triplet in chunks(atoms, 3):  # yield chunks of 3 atom lines each
            if len(atom_triplet) != 3:
                raise ValueError(
                    f"Failed to construct residues. Got only {len(atom_triplet)} ATOM lines"
                )
            # check if all three atoms belong to the same residue by adding a tuple of their residue defining properties to a set
            unique_residues = {
                (atom.residue_names[0], atom.chain_id, atom.residue_number)
                for atom in atom_triplet
            }
            if len(unique_residues) != 1:
                raise ValueError(
                    f"Failed to construct residues. Atoms of different match_mode, chains, residue types or residue numbers {unique_residues} found in Atom triplet"
                )
            residues.append(Residue(atom_triplet))

        return residues

    @property
    def atoms(
        self,
    ) -> Tuple[pyjess.TemplateAtom, pyjess.TemplateAtom, pyjess.TemplateAtom]:
        """
        Get the `tuple` of three `~pyjess.TemplateAtom` describing the residue.
        """
        return self._atoms

    @property
    def residue_name(self) -> str:
        """
        `str`: Get the amino-acid type as three letter code from the first atom.
        """
        return self.atoms[0].residue_names[0]

    @property
    def allowed_residues(self) -> str:
        """
        `str`: Get the allowed residue types as string of single letter codes.
        """
        convert_to_single = {
            "ALA": "A",
            "CYS": "C",
            "ASP": "D",
            "GLU": "E",
            "PHE": "F",
            "GLY": "G",
            "HIS": "H",
            "ILE": "I",
            "LYS": "K",
            "LEU": "L",
            "MET": "M",
            "ASN": "N",
            "PRO": "P",
            "GLN": "Q",
            "ARG": "R",
            "SER": "S",
            "THR": "T",
            "VAL": "V",
            "TRP": "W",
            "TYR": "Y",
            "XXX": "X",
        }
        return "".join(set(convert_to_single[i] for i in self.atoms[0].residue_names))

    @property
    def specific(self) -> int:
        """
        `bool`: Atoms with a match_mode greater 100 are unspecific. <100 is specific.
        """
        return all([atm.match_mode < 100 for atm in self.atoms])

    @property
    def backbone(self) -> bool:
        """
        `bool`: True if an atom may match a backbone atom.

        True if the atom may match backbone atoms.
        Check if the atom has 'ANY' or 'XXX' in its residue_names attribute
        """

        # TODO check what X or XXX means in pyjess!
        return (
            "ANY" in self.atoms[0].residue_names or "XXX" in self.atoms[0].residue_names
        )

    @property
    def residue_number(self) -> int:
        """
        `int`: Get the pdb residue number from the first atom.
        """
        return self.atoms[0].residue_number

    @property
    def chain_id(self) -> str:
        """
        `str`: Get the pdb chain_id from the first atom.
        """
        return self.atoms[0].chain_id

    @property
    def orientation_vector(self) -> Vec3:
        """
        `Vec3`: Calculate the residue orientation vector according to the residue type.
        """
        return self._vec

    @property
    def orientation_vector_indices(self) -> Tuple[int, int]:
        """
        `tuple`: Return the indices of the atoms
        between which the orientation vector was calculated according to the residue type.
        """
        return self._indices


@dataclass(frozen=True)
class AnnotatedResidue(Residue):
    """
    Child class inheriting from `Residue` for M-CSA annotated template residues.

    Attributes:
        is_mutated : `bool` : Wether the residue in the M-CSA reference PDB structure was mutated.
        is_metal_ligand : `str` : Wether the residue in the M-CSA reference PDB structure was metal coordinating.
        roles : `Tuple[str]` : Tuple of EMO codes describing catalytic roles.
        reference_idx: `int` : Residue indentifier of the reference residue.
        has_ptm: `bool` : Wether the residue in the M-CSA reference PDB structure was post-translationally modified.

    """

    is_mutated: bool
    is_metal_ligand: bool
    roles: Tuple[str, ...]
    reference_idx: int
    has_ptm: bool


@dataclass(frozen=True)
class Cluster:
    """
    Class for storing template cluster information.

    Attributes:
        id: `int` Index of the template cluster
        member: `int` Member index within the template cluster
        size: `int` Total number of members in the template cluster
    """

    id: int
    member: int
    size: int

    def __post_init__(self):
        if self.member > self.size:
            raise ValueError("Cluster member cannot be greater than cluster size")


class Template(pyjess.Template):
    """
    Class for storing templates and associated information.

    Inherits and extends from `pyjess.Template`
    """

    _CATH_MAPPING: ClassVar[Dict[str, List[str]]]
    _EC_MAPPING: ClassVar[Dict[str, str]]
    _PDB_SIFTS: ClassVar[Dict[str, Dict[str, List[str]] | str]]

    def __init__(
        self,
        residues: Sequence[Residue],
        pdb_id: Optional[str] = None,
        mcsa_id: Optional[int] = None,
        *,
        id: Optional[str] = None,
        template_id_string: Optional[str] = None,
        cluster: Optional[Cluster] = None,
        uniprot_id: Optional[str] = None,
        organism: Optional[str] = None,
        organism_id: Optional[str] = None,
        resolution: Optional[float] = None,
        experimental_method: Optional[str] = None,
        enzyme_discription: Optional[str] = None,
        represented_sites: Optional[int] = None,
        ec: Iterable[str] = (),
        cath: Iterable[str] = (),
    ):
        """
        Initialize a template.

        Keyword Arguments:
            residues: `sequence` of `~Residue` instances
            id: `str` Internal Template ID string. Default `None`
            pdb_id: `str` The PDB ID of the template
            template_id_string: `str` String in the ID line of a template
            mcsa_id: `int` The M-CSA entry index form which the template was generated
            cluster: `Cluster` Instance of the template
            uniprot_id: `str` UniProt Identifier of the Protein from which the template was generated
            organism: `str` Organism name of the Protein from which the template was generated
            organism_id: `str` Taxonomic Identifier of the Organism of the Protein from which the template was generated
            resolution: `float` Resolution of the Protein Structure from which the template was generated
            experimental_method: `str` Experimental method by which the Protein Structure of the template was resolved
            enzyme_description: `str` Text Discription of the Protein from which the template was generated
            represented_sites: `int` The number of Enzymes which this template is representative for
            ec: `list` of EC numbers associated with Enzymes this template represents
            cath: `list` of CATH numbers associated with Enzymes this template represents

        NOTE:
            It is recommended not to pass an id string.
            If id is `None`, the id string will be set to:

            > {`~Template.effective_size`}-Residues_{~`Template.template_id_string`}_Cluster_{`Cluster.id`}-{`Cluster.member`}-{`Cluster.size`}

            This identifier string should be unique.

        NOTE:
            residues can be constructed from a list of ~pyjess.TemplateAtoms via the
            staticmethod ~Residue.construct_residues_from_atoms(atoms=atoms)

        Returns:
            `Template`
        """
        if len(residues) == 0:
            raise ValueError(
                "Tried creating an `Template` from an empty list of residues!"
            )

        for residue in residues:
            if not isinstance(residue, type(residues[0])):
                raise ValueError(
                    f"Tried creating a `Template` from different types of `Residue` objects. Got {type(residue) and type(residues[0])}"
                )

        atoms: List[pyjess.TemplateAtom] = []
        for residue in residues:
            atoms.extend(residue._atoms)

        super().__init__(atoms, id=id)

        self.residues = tuple(residues)
        self.pdb_id = pdb_id
        self.mcsa_id = mcsa_id
        self.template_id_string = template_id_string
        self.cluster = cluster
        self.uniprot_id = uniprot_id
        self.organism = organism
        self.organism_id = organism_id
        self.resolution = resolution
        self.experimental_method = experimental_method
        self.enzyme_discription = enzyme_discription
        self.represented_sites = represented_sites
        self.enzyme_discription = enzyme_discription
        self.ec = tuple(sorted({*ec, *self._add_ec_annotations()}))
        self.cath = tuple(sorted({*cath, *self._add_cath_annotations()}))

    def _state(self) -> Tuple:
        return (
            tuple(self.residues),
            self.id,
            self.pdb_id,
            self.mcsa_id,
            self.template_id_string,
            self.cluster,
            self.effective_size,
            self.uniprot_id,
            self.organism,
            self.organism_id,
            self.resolution,
            self.experimental_method,
            self.enzyme_discription,
            self.represented_sites,
            self.ec,
            self.cath,
        )

    def __copy__(self) -> Template:
        return self.copy()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Template):
            self_state = self._state()
            other_state = other._state()
            return self_state == other_state
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, Template):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self) -> int:
        return hash((type(self), self._state()))

    @classmethod  # reading from text
    def loads(
        cls,
        text: str,
        id: str | None = None,
        warn: bool = False,
    ) -> Template:
        """
        Load Template from `str`. Calls `Template.load()`

        Arguments:
            text: `str` of Template to load
            id: `str` or `None` Internal pyjess string which will superseed the ID string parsed from the template file. Default `None`
            warn: `bool` If warnings should be printed. Default `False`

        Returns:
            `Template`
        """
        return cls.load(io.StringIO(text), id=id, warn=warn)

    @classmethod  # reading from TextIO
    def load(
        cls,
        file: TextIO | Iterator[str] | str | os.PathLike[str],
        id: str | None = None,
        warn: bool = False,
    ) -> Template:
        """
        Overloaded load to parse a `pyjess.Template` and its associated info into an `Template` object

        Arguments:
            file: `file-like` object or `str` or `path-like` from which to load
            id: `str` or `None` Internal pyjess string which will superseed the ID string parsed from the template file. Default `None`
            warn: `bool` If warnings should be printed. Default `False`

        Returns:
            `Template`
        """
        atoms = []
        metadata: dict[str, object] = {"ec": list(), "cath": list()}

        _PARSERS = {
            "ID": cls._parse_template_id_string,
            "PDB_ID": cls._parse_pdb_id,
            "UNIPROT_ID": cls._parse_uniprot_id,
            "MCSA_ID": cls._parse_mcsa_id,
            "CLUSTER": cls._parse_cluster,
            "ORGANISM_NAME": cls._parse_organism_name,
            "ORGANISM_ID": cls._parse_organism_id,
            "RESOLUTION": cls._parse_resolution,
            "EXPERIMENTAL_METHOD": cls._parse_experimental_method,
            "EC": cls._parse_ec,
            "CATH": cls._parse_cath,
            "ENZYME": cls._parse_enzyme_discription,
            "REPRESENTING": cls._parse_represented_sites,
        }

        # NOTE
        # currently mutliple templates per file are not supported by this parser function.

        # NOTE
        # this parser does not check for the existance of a header like REMARK TEMPLATE

        try:
            context: ContextManager[TextIO] = open(os.fspath(file))  # type: ignore
        except TypeError:
            context = contextlib.nullcontext(file)  # type: ignore

        with context as f:
            seen_lines = set()
            for line in filter(str.strip, f):
                tokens = line.split()
                if tokens[0] == "REMARK":
                    if len(tokens) == 1:
                        continue  # skip lines which just REMARK and nothing else
                    parser = _PARSERS.get(tokens[1])
                    if parser is not None:
                        if len(tokens) < 3:
                            raise IndexError(
                                f"Expected some annotation after the REMARK {tokens[1]} flag"
                            )
                        elif tokens[2].upper() in ["NONE", "?", "NAN", "NA"]:
                            continue
                        parser(tokens, metadata, warn=warn)
                elif tokens[0] == "ATOM":
                    if line in seen_lines:
                        raise ValueError("Duplicate Atom lines passed!")
                    seen_lines.add(line)
                    atoms.append(pyjess.TemplateAtom.loads(line))
                elif tokens[0] == "HETATM":
                    raise ValueError(
                        "Supplied template with HETATM record. HETATMs cannot be searched by Jess"
                    )
                else:
                    continue

        residues = Residue.construct_residues_from_atoms(atoms=atoms)

        return Template(
            residues=residues,
            id=id,
            **metadata,  # type: ignore # unpack everything parsed into metadata
        )

    def copy(self) -> Template:
        return type(self)(
            residues=self.residues,
            id=self.id,
            pdb_id=self.pdb_id,
            template_id_string=self.template_id_string,
            mcsa_id=self.mcsa_id,
            cluster=self.cluster,
            uniprot_id=self.uniprot_id,
            organism=self.organism,
            organism_id=self.organism_id,
            resolution=self.resolution,
            experimental_method=self.experimental_method,
            enzyme_discription=self.enzyme_discription,
            represented_sites=self.represented_sites,
            ec=self.ec,
            cath=self.cath,
        )

    def dumps(self) -> str:
        # """
        # Dump `AnnotatedTemplate` to a `str`. Calls `AnnotatedTemplate.dump()`
        # """
        buffer = io.StringIO()
        self.dump(buffer)
        return (
            buffer.getvalue()
        )  # returns entire content temporary file object as a string

    def dump(self, file: TextIO):
        # """
        # Dump `AnnotatedTemplate` to `file-like` object.

        # Arguments:
        #     file: `file-like` object to write to
        # """

        raise NotImplementedError

        file.write("REMARK TEMPLATE\n")
        if self.template_id_string:
            file.write(f"REMARK ID {self.template_id_string}\n")
        if self.pdb_id:
            file.write(f"REMARK PDB_ID {self.pdb_id}\n")
        if self.mcsa_id:
            file.write(f"REMARK MCSA_ID {self.mcsa_id}\n")
        if self.uniprot_id:
            file.write(f"REMARK UNIPROT_ID {self.uniprot_id}\n")
        if self.cluster:
            file.write(
                f"REMARK CLUSTER {'_'.join([str(self.cluster.id), str(self.cluster.member), str(self.cluster.size)])}\n"
            )
        if self.organism:
            file.write(f"REMARK ORGANISM_NAME {self.organism}\n")
        if self.organism_id:
            file.write(f"REMARK ORGANISM_ID {self.organism_id}\n")
        if self.resolution:
            file.write(f"REMARK RESOLUTION {self.resolution}\n")
        if self.experimental_method:
            file.write(f"REMARK EXPERIMENTAL_METHOD {self.experimental_method}\n")
        if self.enzyme_discription:
            file.write(f"REMARK ENZYME {self.enzyme_discription}\n")
        if self.represented_sites:
            file.write(
                f"REMARK REPRESENTING {self.represented_sites} CATALYTIC SITES\n"
            )
        if self.experimental_method:
            file.write(f"REMARK EXPERIMENTAL_METHOD {self.experimental_method}\n")
        if self.ec:
            file.write(f"REMARK EC {','.join(self.ec)}\n")
        if self.cath:
            file.write(f"REMARK CATH {','.join(self.cath)}")
        # file.write(f"REMARK SIZE {self.effective_size}\n")
        # file.write(f"REMARK DIMENSION {self.dimension}\n")
        # file.write(f"REMARK MULTIMERIC {self.multimeric}\n")
        # file.write(f"REMARK RELATIVE_ORDER {self.relative_order}\n")

        # for residue in self.residues:
        #     file.write(
        #         f"REMARK ORIENTATION_VECTOR OF RESIDUE {residue.residue_number}: between atom {residue.orientation_vector_indices[0]} and {residue.orientation_vector_indices[1]} {residue.orientation_vector.x:.3f} {residue.orientation_vector.y:.3f} {residue.orientation_vector.z:.3f}\n"
        #     )

        for residue in self.residues:
            for atom in residue.atoms:
                raise NotImplementedError(
                    "Still need to write a dump method for templates in PyJess"
                )
                # TODO atom.dump(file)

        file.write("END\n")

    @cached_property
    def effective_size(self) -> int:
        """`int`: The number of unique residues in the template, excluding backbone residues and unspecific residues."""
        # Number of residues as evaluated, the effective size
        # Effective size of a template is not necessarily equal to the number of atom triplets in a template:
        # Residues matching ANY amino acid type are not counted as they are too general
        # These have a value of 100 or hgiher in the second column indicating unspecific residues
        # Not all BB residues are unspecific! Some are targeted towards only Gly for example and thus have values < 100

        # Even if a Residue must match a particular AA, 6 atoms from a given residue may be selected
        # once for main chain and once for side chain
        # Therefore we only count unique template residues!

        # It seems that some templates even have the same 3-atom triplets at the same location twice. This I assume must be an error
        # again a reason to only count unique residues

        effective_size = 0
        for residue in self.residues:
            if (
                residue.specific and not residue.backbone
            ):  # type specific and not Backbone
                effective_size += 1
        return effective_size

    @cached_property
    def multimeric(
        self,
    ) -> bool:  # if the template is split across multiple protein chains
        """`bool`: True if the template contains residues from multiple protein chains."""
        return not all(
            res.chain_id == list(self.residues)[0].chain_id for res in self.residues
        )

    # TODO add annotation for homo-meric or hetero-meric!
    # these would have to be m-csa derived. in the entry json,
    # such catalytic sites are labeled under reaction as polymeric.
    # homoeric multimeric sites are not labeled polymeric

    @cached_property
    def relative_order(
        self,
    ) -> List[int]:  # list with length of deduplicated template dimension
        """`list` of `int`: Relative order of residues in the template sorted by the pdb residue number.

        Note:
            This only works for non-multimeric templates. In this case returns '[0]'.
        """
        if self.multimeric:
            return [0]
        else:
            # Now extract relative template order
            return ranked_argsort([res.residue_number for res in self.residues])

    def _add_cath_annotations(self) -> List[str]:
        """`list`: Pull CATH Ids associated with that template from SIFTS and from the M-CSA"""
        cath_list = []
        if self.mcsa_id:
            cath_list.extend(
                self._CATH_MAPPING[str(self.mcsa_id)]
            )  # this is a bit inaccurate possibly ... shouldnt we pull via pdb_id....
        if self.pdb_id:
            pdbchains = set()
            for res in self.residues:
                pdbchains.add("{}{}".format(self.pdb_id, res.chain_id))

            # Iterating of all pdbchains which were part of the AnnotatedTemplate
            for pdbchain in pdbchains:
                # also include CATH annotations from PDB-SIFTS
                subdict = self._PDB_SIFTS.get(pdbchain)
                if subdict is not None:
                    sifts_caths = subdict.get("cath").copy()  # type: ignore
                    for cath in sifts_caths:
                        if cath != "?":
                            cath_list.append(cath)

        return cath_list

    def _add_ec_annotations(self) -> List[str]:
        """`list`: Pull EC Annotations associated with that template from SIFTS and from the M-CSA"""
        ec_list = []
        if self.mcsa_id is not None:
            ec_list.append(
                self._EC_MAPPING[str(self.mcsa_id)]
            )  # this is a bit inaccurate possibly ... shouldnt we pull via pdb_id....
        if self.pdb_id is not None:
            pdbchains = set()
            for res in self.residues:
                pdbchains.add("{}{}".format(self.pdb_id, res.chain_id))

            # Iterating of all pdbchains which were part of the AnnotatedTemplate
            for pdbchain in pdbchains:
                # also include EC annotations from PDB-SIFTS
                subdict = self._PDB_SIFTS.get(pdbchain)
                if subdict is not None:
                    sifts_ecs = subdict.get("ec").copy()  # type: ignore
                    for ec in sifts_ecs:
                        if ec != "?":
                            ec_list.append(ec)

        return ec_list

    @classmethod
    def _parse_pdb_id(
        cls, tokens: List[str], metadata: dict[str, object], warn: bool = True
    ):
        if len(tokens[2]) != 4:
            raise ValueError(
                f"Found {tokens[2]} which has more than the expected 4 characters of a PDB_ID."
            )
        metadata["pdb_id"] = tokens[2].lower()

    @classmethod
    def _parse_template_id_string(
        cls, tokens: List[str], metadata: dict[str, object], warn: bool = True
    ):
        metadata["template_id_string"] = tokens[2]

    @classmethod
    def _parse_uniprot_id(
        cls, tokens: List[str], metadata: dict[str, object], warn: bool = True
    ):
        match = re.search(
            r"[OPQ][0-9][A-Z0-9]{3}[0-9]|[A-NR-Z][0-9]([A-Z][A-Z0-9]{2}[0-9]){1,2}",
            tokens[2],
        )
        if match:
            metadata["uniprot_id"] = match.group()
        else:
            raise ValueError(f"Did not find a valid UniProt ID, found {tokens[2]}")

    @classmethod
    def _parse_mcsa_id(
        cls, tokens: List[str], metadata: dict[str, object], warn: bool = True
    ):
        try:
            metadata["mcsa_id"] = int(tokens[2])
        except ValueError as exc:
            raise ValueError(f"Did not find a M-CSA ID, found {tokens[2]}") from exc

    @classmethod
    def _parse_cluster(
        cls, tokens: List[str], metadata: dict[str, object], warn: bool = True
    ):
        try:
            cluster = Cluster(*list(map(int, tokens[2].split("_"))))
            metadata["cluster"] = cluster
        except ValueError as exc:
            raise ValueError(
                f"Did not find a Cluster specification in the form <id>_<member>_<size>, found {tokens[2]}"
            ) from exc

    @classmethod
    def _parse_organism_name(
        cls, tokens: List[str], metadata: dict[str, object], warn: bool = True
    ):
        metadata["organism"] = " ".join(tokens[2:])

    @classmethod
    def _parse_organism_id(
        cls, tokens: List[str], metadata: dict[str, object], warn: bool = True
    ):
        metadata["organism_id"] = tokens[2]

    @classmethod
    def _parse_resolution(
        cls, tokens: List[str], metadata: dict[str, object], warn: bool = True
    ):
        try:
            metadata["resolution"] = float(tokens[2])
        except ValueError as exc:
            raise ValueError(f"Ill-formatted pdb resolution: {tokens[2]}") from exc

    @classmethod
    def _parse_experimental_method(
        cls, tokens: List[str], metadata: dict[str, object], warn: bool = True
    ):
        metadata["experimental_method"] = " ".join(tokens[2:])

    @classmethod
    def _parse_ec(
        cls, tokens: List[str], metadata: dict[str, object], warn: bool = True
    ):
        matches = [
            match.group()
            for match in re.finditer(r"[1-7](\.(\-|\d{1,})){3}", tokens[2])
        ]
        non_cat_matches = [
            match.group()
            for match in re.finditer(r"[1-7](\.(\-|\d{1,}|n\d{1,})){3}", tokens[2])
        ]
        if matches:
            for ec in matches:
                if ec not in metadata["ec"]:  # type: ignore
                    metadata["ec"].append(ec)  # type: ignore
        elif non_cat_matches:
            for ec in non_cat_matches:
                if ec not in metadata["ec"]:  # type: ignore
                    metadata["ec"].append(ec)  # type: ignore
            if warn:
                warnings.warn(
                    f"Rare EC number(s) {[ec for ec in non_cat_matches]} presumed to be noncatalytic detected!"
                )
        else:
            raise ValueError(f"Did not find a valid EC number, found {tokens[2]}")

    @classmethod
    def _parse_cath(
        cls, tokens: List[str], metadata: dict[str, object], warn: bool = True
    ):
        matches = [
            match.group()
            for match in re.finditer(r"[1-46](\.(\-|\d{1,})){3}", tokens[2])
        ]
        if matches:
            for cath in matches:
                if cath not in metadata["cath"]:  # type: ignore
                    metadata["cath"].append(cath)  # type: ignore
        else:
            raise ValueError(f"Did not find a valid CATH number, found {tokens[2]}")

    @classmethod
    def _parse_enzyme_discription(
        cls, tokens: List[str], metadata: dict[str, object], warn: bool = True
    ):
        metadata["enzyme_discription"] = " ".join(tokens[2:])

    @classmethod
    def _parse_represented_sites(
        cls, tokens: List[str], metadata: dict[str, object], warn: bool = True
    ):
        try:
            metadata["represented_sites"] = int(tokens[2])
        except ValueError as exc:
            raise ValueError(
                f"Ill-formatted number of represented sites: {tokens[2]}"
            ) from exc


class AnnotatedTemplate(Template):

    residues: Tuple[AnnotatedResidue, ...]
    pdb_id: str
    mcsa_id: int

    def __init__(
        self,
        residues: Sequence[AnnotatedResidue],
        pdb_id: str,
        mcsa_id: int,
        *,
        number_of_mutated_residues: int,
        number_of_metal_ligands: Tuple[int, int],
        number_of_ptm_residues: Tuple[int, int],
        number_of_side_chain_residues: Tuple[int, int],
        total_reference_residues: int,
        assembly: int,
        id: Optional[str] = None,
        template_id_string: Optional[str] = None,
        cluster: Optional[Cluster] = None,
        uniprot_id: Optional[str] = None,
        organism: Optional[str] = None,
        organism_id: Optional[str] = None,
        resolution: Optional[float] = None,
        experimental_method: Optional[str] = None,
        enzyme_discription: Optional[str] = None,
        represented_sites: Optional[int] = None,
        ec: Iterable[str] = (),
        cath: Iterable[str] = (),
    ):
        """
        Initialize an annotated template with descriptions of catalytic activity from the M-CSA.

        Keyword Arguments:
            residues: `sequence` of `~Residue` instances
            id: `str` Internal Template ID string. Default `None`
            pdb_id: `str` The PDB ID of the template
            template_id_string: `str` String in the ID line of a template
            mcsa_id: `int` The M-CSA entry index form which the template was generated
            cluster: `Cluster` Instance of the template
            uniprot_id: `str` UniProt Identifier of the Protein from which the template was generated
            organism: `str` Organism name of the Protein from which the template was generated
            organism_id: `str` Taxonomic Identifier of the Organism of the Protein from which the template was generated
            resolution: `float` Resolution of the Protein Structure from which the template was generated
            experimental_method: `str` Experimental method by which the Protein Structure of the template was resolved
            enzyme_description: `str` Text Discription of the Protein from which the template was generated
            represented_sites: `int` The number of Enzymes which this template is representative for
            ec: `list` of EC numbers associated with Enzymes this template represents
            cath: `list` of CATH numbers associated with Enzymes this template represents
            number_of_mutated_residues: `int` The number of side chain specific residues which have been mutated relative to the reference
            number_of_metal_ligands: `tuple(`int`, `int`)` Number of metal chelating residues in the template and the reference
            number_of_ptm_residues: `tuple(`int`, `int`)` Number of post translationally modified residues in the template and the reference
            number_of_side_chain_residues: `tuple(`int`, `int`)` Number of side chain interacting residues in the template and the reference
            total_reference_residues: `int` Total number of residues (main and side chain) in the reference structure

        NOTE:
            In order for a template file to be loaded as an AnnotatedTemplate,
            it must need both an M-CSA id and a pdb_id.
            This pdb_id must be found in the PDB-homologs of the M-CSA!

        NOTE:
            It is recommended not to pass an id string.
            If id is `None`, the id string will be set to

            > {`~Template.effective_size`}-Residues_{~`Template.template_id_string`}_Cluster_{`Cluster.id`}-{`Cluster.member`}-{`Cluster.size`}

            This identifier string should be unique.

        Returns:
            `AnnotatedTemplate`
        """

        # if mcsa_id is None or pdb_id is None:
        #     raise ValueError("Missing mcsa_id or pdb_id for AnnotatedTemplate object.")

        super().__init__(
            residues=residues,
            pdb_id=pdb_id,
            mcsa_id=mcsa_id,
            id=id,
            template_id_string=template_id_string,
            cluster=cluster,
            uniprot_id=uniprot_id,
            organism=organism,
            organism_id=organism_id,
            resolution=resolution,
            experimental_method=experimental_method,
            enzyme_discription=enzyme_discription,
            represented_sites=represented_sites,
            ec=ec,
            cath=cath,
        )

        # Now add in all the special information
        self.residues = tuple(residues)
        self.number_of_mutated_residues = number_of_mutated_residues
        self.number_of_metal_ligands = number_of_metal_ligands
        self.number_of_ptm_residues = number_of_ptm_residues
        self.number_of_side_chain_residues = number_of_side_chain_residues
        self.total_reference_residues = total_reference_residues
        self.assembly = assembly

    def _state(self) -> Tuple:
        atoms: List[pyjess.TemplateAtom] = []
        for residue in self.residues:
            atoms.extend(residue._atoms)

        residues = Residue.construct_residues_from_atoms(atoms=atoms)
        # We use reuglar unanntoated residues to get the state!
        return (
            tuple(residues),
            self.id,
            self.pdb_id,
            self.mcsa_id,
            self.template_id_string,
            self.cluster,
            self.effective_size,
            self.uniprot_id,
            self.organism,
            self.organism_id,
            self.resolution,
            self.experimental_method,
            self.enzyme_discription,
            self.represented_sites,
            self.ec,
            self.cath,
        )

    def __copy__(self) -> AnnotatedTemplate:
        return self.copy()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, AnnotatedTemplate):
            self_state = self._state()
            other_state = other._state()
            return self_state == other_state
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, AnnotatedTemplate):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self) -> int:
        return hash((type(self), self._state()))

    def copy(self) -> AnnotatedTemplate:
        return type(self)(
            residues=self.residues,
            id=self.id,
            pdb_id=self.pdb_id,
            template_id_string=self.template_id_string,
            mcsa_id=self.mcsa_id,
            cluster=self.cluster,
            uniprot_id=self.uniprot_id,
            organism=self.organism,
            organism_id=self.organism_id,
            resolution=self.resolution,
            experimental_method=self.experimental_method,
            enzyme_discription=self.enzyme_discription,
            represented_sites=self.represented_sites,
            ec=self.ec,
            cath=self.cath,
            number_of_mutated_residues=self.number_of_mutated_residues,
            number_of_metal_ligands=self.number_of_metal_ligands,
            number_of_ptm_residues=self.number_of_ptm_residues,
            number_of_side_chain_residues=self.number_of_side_chain_residues,
            total_reference_residues=self.total_reference_residues,
            assembly=self.assembly,
        )

    @classmethod  # reading from text
    def loads(
        cls,
        text: str,
        id: str | None = None,
        warn: bool = False,
        with_annotations: bool = True,
    ) -> Template | AnnotatedTemplate:
        """
        Load Template from `str`. Calls `Template.load()`

        Arguments:
            text: `str` of Template to load
            id: `str` or `None` Internal pyjess string which will superseed the ID string parsed from the template file. Default `None`
            warn: `bool` If warnings should be printed. Default `False`
            with_annotations: `bool` If True (default) M-CSA derived templates with a PDB-id and M-CSA id will be annotated with extra information.

        Returns:
            `Template|AnnotatedTemplate`
        """
        return cls.load(
            io.StringIO(text), id=id, warn=warn, with_annotations=with_annotations
        )

    @classmethod  # reading from TextIO
    def load(
        cls,
        file: TextIO | Iterator[str] | str | os.PathLike[str],
        id: str | None = None,
        warn: bool = False,
        with_annotations: bool = True,
    ) -> Template | AnnotatedTemplate:
        """
        Overloaded load to parse a `pyjess.Template` and its associated info into an `Template` object

        Arguments:
            file: `file-like` object or `str` or `path-like` from which to load
            id: `str` or `None` Internal pyjess string which will superseed the ID string parsed from the template file. Default `None`
            warn: `bool` If warnings should be printed. Default `False`
            with_annotations: `bool` If True (default) M-CSA derived templates with a PDB-id and M-CSA id will be annotated with extra information.

        Returns:
            `Template`|`AnnotatedTemplate`
        """

        template = Template.load(
            file=file,
            id=id,
            warn=warn,
        )

        if (
            with_annotations
            and template.pdb_id is not None
            and template.mcsa_id is not None
        ):
            annotated_residues, ann_dict = cls.derive_mcsa_annotations(template)

            return AnnotatedTemplate(
                residues=annotated_residues,
                id=template.id,
                pdb_id=template.pdb_id,
                mcsa_id=template.mcsa_id,
                template_id_string=template.template_id_string,
                cluster=template.cluster,
                uniprot_id=template.uniprot_id,
                organism=template.organism,
                organism_id=template.organism_id,
                resolution=template.resolution,
                experimental_method=template.experimental_method,
                enzyme_discription=template.enzyme_discription,
                represented_sites=template.represented_sites,
                ec=template.ec,
                cath=template.cath,
                number_of_mutated_residues=ann_dict["number_of_mutated_residues"],
                number_of_metal_ligands=ann_dict["number_of_metal_ligands"],
                number_of_ptm_residues=ann_dict["number_of_ptm_residues"],
                number_of_side_chain_residues=ann_dict["number_of_side_chain_residues"],
                total_reference_residues=ann_dict["total_reference_residues"],
                assembly=ann_dict["assembly_id"],
            )

        else:
            return template

    @staticmethod
    def derive_mcsa_annotations(
        template: Template,
    ) -> Tuple[List[AnnotatedResidue], Dict[str, Any]]:

        # NOTE
        # be careful how to interpret and handle this data.
        # Some templates are build from the reference pdb structure in the entry
        # Others, are built from a non-reference structure
        # Further, there are multimeric mcsa catalytic sites composed of:
        #   homo-mers (Example M-CSA 1, 4, ...)
        #   hetero-mers (Example M-CSA 5, 10, 11, ...)

        # NOTE
        # This only works if the template comes from the M-CSA (with id) and PDB id

        if template.mcsa_id is None or template.pdb_id is None:
            raise ValueError(
                "Tried annotating a template lacking either pdb_id or mcsa_id"
            )

        reference_homologs = set()
        annotated_residues = []
        assembly_ids = set()
        for residue in template.residues:
            ################### get reference residue ##################################
            try:
                template_pdbchain = CATALYTIC_RESIDUE_HOMOLOGS[template.mcsa_id][  # type: ignore
                    template.pdb_id + residue.chain_id  # type: ignore
                ]
            except KeyError:
                try:
                    template_pdbchain = CATALYTIC_RESIDUE_HOMOLOGS[template.mcsa_id][  # type: ignore
                        template.pdb_id + residue.chain_id[0]  # type: ignore
                    ]
                    # print(f"Only found a template pdb for {template.pdb_id} with pdbchain {residue.chain_id[0]} instead of {residue.chain_id}")
                    # these are all homo-mers except for 1olx
                    # but which is correctly assigned here too
                except KeyError:
                    raise KeyError(
                        f"Failed to find template pdb in catalytic residue homologs for M-CSA id {template.mcsa_id} and pdbchain {template.pdb_id+residue.chain_id}"  # type: ignore
                    ) from None

            assembly_ids.add(template_pdbchain.assembly)

            match_found = False
            for index, hom_residue in template_pdbchain.residues.items():
                # TODO check if it would even make a difference
                # TODO Ask ioannis about this: Should I check auth_resid or resid first?
                if residue.residue_number == hom_residue.auth_resid:
                    match_found = True
                    break
                elif residue.residue_number == hom_residue.resid:
                    match_found = True
                    break

            if not match_found:
                raise ValueError(
                    f"Missing a comparison residue for M-CSA id {template.mcsa_id} and pdbchain {template.pdb_id+residue.chain_id} for residue {residue.residue_name, residue.residue_number}"  # type: ignore
                ) from None

            # after the for loop breaks, index contains the residue index
            # and hom_residue contains the corresponding residue annotations
            if isinstance(hom_residue, ReferenceCatalyticResidue):
                reference_pdb = template_pdbchain
                ref_residue = hom_residue
            elif isinstance(hom_residue, NonReferenceCatalyticResidue):
                _, ref_pdbchain, _ = hom_residue.reference
                reference_pdb = CATALYTIC_RESIDUE_HOMOLOGS[template.mcsa_id][  # type: ignore
                    ref_pdbchain
                ]
                ref_residue = reference_pdb.residues[index]  # type: ignore

            reference_homologs.add(reference_pdb)

            ############################################################################

            # We check mutation only if
            # the type of the template residue matters
            # and it is interacting via its sidechain
            is_mutated = False
            if (
                residue.residue_name not in ["ANY", "PTM"]
                and residue.specific
                and ref_residue.function_location_abv is not None
            ):
                is_mutated = residue.residue_name != ref_residue.code

            # if template.mcsa_id == 661 and template.pdb_id == "4cyr" and ref_residue.resid == 51:
            #     print(ref_pdbchain, ref_residue.resid, ref_residue.code, ref_residue.ptm, bool(ref_residue.ptm))

            #     if isinstance(hom_residue, NonReferenceCatalyticResidue):
            #         print("is non_reference")

            annotated_residues.append(
                AnnotatedResidue(
                    _atoms=residue._atoms,
                    _vec=residue._vec,
                    _indices=residue._indices,
                    reference_idx=index,
                    is_metal_ligand="metal ligand" in ref_residue.roles_summary,
                    is_mutated=is_mutated,
                    has_ptm=bool(ref_residue.ptm),
                    roles=tuple(ref_residue.roles),
                )
            )

        if len(assembly_ids) == 1:
            assembly_id = list(assembly_ids)[0]
        else:
            raise ValueError(
                f"Got multiple assemblies for template from {template.pdb_id}"
            )

        ######################### Template level counts ################################
        # for templates where all the residues have annotations
        # we loop again over these residues to get template level counts
        number_side_chain_residues = 0
        number_metal_ligands = 0
        number_ptm = 0
        number_mutated = 0

        for residue in annotated_residues:
            if residue.specific:
                number_side_chain_residues += 1
            if residue.is_metal_ligand:
                number_metal_ligands += 1
            if residue.has_ptm:
                number_ptm += 1
            if residue.is_mutated:
                number_mutated += 1

        ################################################################################

        ############### Reference Homolog counts #######################################
        number_reference_residues = 0
        number_side_chain_reference_residues = 0
        number_metal_ligands_reference = 0
        number_ptm_reference = 0
        for reference_pdb in reference_homologs:
            for residue_id, residue in reference_pdb.residues.items():  # type: ignore
                if isinstance(residue, ReferenceCatalyticResidue):
                    # skip if the residue doesnt exist in the reference
                    if residue.code is None:
                        continue
                    number_reference_residues += 1
                    # function_location_abv is only set if it is NOT a side chain interaction
                    if not residue.function_location_abv:
                        number_side_chain_reference_residues += 1
                    if "metal ligand" in residue.roles_summary:
                        number_metal_ligands_reference += 1
                    if residue.ptm:
                        number_ptm_reference += 1

        ######################## assing Template properties/attributes #################

        return annotated_residues, {
            # number of mutated residues in the template versus the reference.
            "number_of_mutated_residues": number_mutated,
            # tuple(template, reference)
            "number_of_metal_ligands": (
                number_metal_ligands,
                number_metal_ligands_reference,
            ),
            # tuple(template, reference)
            "number_of_ptm_residues": (number_ptm, number_ptm_reference),
            # Side chain residues and ANY and PTM are discarded
            # tuple(template, reference)
            "number_of_side_chain_residues": (
                number_side_chain_residues,
                number_side_chain_reference_residues,
            ),
            # total number of residues in the reference pdb structure
            "total_reference_residues": number_reference_residues,
            "assembly_id": assembly_id,
        }


# Populate the mapping of MCSA IDs to CATH numbers so that it can be accessed
# by individual templates in the `Template.cath` property.
# Source: M-CSA which provides cath annotations for either residue homologs or for m-csa entries
with resource_files(__package__).joinpath("data/MCSA_CATH_mapping.json").open() as f:
    Template._CATH_MAPPING = json.load(f)

with resource_files(__package__).joinpath("data/MCSA_EC_mapping.json").open() as f:
    Template._EC_MAPPING = json.load(f)

# Source: CATH, EC and InterPro from PDB-SIFTS through mapping to the pdbchain
with resource_files(__package__).joinpath("data/pdb_sifts.json").open() as f:
    Template._PDB_SIFTS = json.load(f)

CATALYTIC_RESIDUE_HOMOLOGS = load_mcsa_catalytic_residue_homologs_info(
    resource_files(__package__).joinpath("data")  # type: ignore
)

# global MCSA_interpro_dict
# # dictonariy mapping M-CSA entries to Interpro Identifiers
# # Interpro Acceccesions at the Domain, Family and Superfamily level
# # are searched for the reference sequences of each M-CSA entry.
# # Note that an M-CSA entry may have multiple reference sequences

# TODO
# # add a list of cofactors associated with each EC number from Neeras List
# cofactors = set()
# if Template_EC in cofactor_dict:
#     cofactors.update(cofactor_dict[Template_EC])


# Find all template files which end in .pdb
def _get_paths_by_extension(directory_path: Path, extension: str) -> List[Path]:
    pattern = f"{directory_path}/**/*{extension}"
    file_paths = glob.glob(pattern, recursive=True)
    files = []
    for file in file_paths:
        files.append(Path(file))

    if files:
        return files
    else:
        raise FileNotFoundError(
            f"No template files with the {extension} extension found in the {directory_path.resolve()} directory"
        )


def load_templates(
    template_dir: Path | None,
    warn: bool = False,
    verbose: bool = False,
    cpus: int = (
        len(os.sched_getaffinity(0)) if sys.platform == "linux" else os.cpu_count() or 1
    ),
    with_annotations: bool = True,
) -> Iterator[Template | AnnotatedTemplate]:
    """
    Load templates from a given directory, recursively.

    Arguments:
        template_dir: `~pathlib.Path` | `None` Directory which to search recursively for files with the '.pdb' extension. If `None`, defaults to templates included in this library.
        warn: `bool` If warnings about annoation issues in templates should be printed. Default `False`
        verbose: `bool` If loading should be verbose. Default `False`
        cpus: `int` The number of CPU threads to use. If (default), use all threads. If <0, leave this number of threads free.
        with_annotations: `bool` If True (default) M-CSA derived templates with a PDB-id and M-CSA id will be annotated with extra information.

    Yields:
        `Template`|`AnnotatedTemplate`
    """
    if template_dir is None:
        template_dir = Path(
            str(resource_files(__package__).joinpath("jess_templates_20230210"))
        )

    elif isinstance(template_dir, Path):
        if not template_dir.is_dir():
            raise NotADirectoryError(
                f"The path {template_dir} doesnt exist or is not a directory!"
            )

    if verbose:
        print(f"Loading Template files from {str(template_dir.resolve())}")

    template_paths = _get_paths_by_extension(template_dir, ".pdb")

    def _load_and_annotate(template_path: Path, warn: bool, with_annotations: bool):
        try:
            with template_path.open() as f:
                return AnnotatedTemplate.load(
                    file=f, warn=warn, with_annotations=with_annotations
                )
        except ValueError as exc:
            raise ValueError(
                f"Passed Template file {template_path.resolve()} contained ATOM lines which are not in Jess Template format."
            ) from exc
        except KeyError as exc:
            raise ValueError(
                f"Passed Template file {template_path.resolve()} contained issues with some residues."
            ) from exc

    if cpus <= 0:
        os_cpu_count = (
            len(os.sched_getaffinity(0)) if sys.platform == "linux" else os.cpu_count()
        )
        if os_cpu_count is not None:
            cpus = max(1, os_cpu_count + cpus)
        else:
            cpus = 1

    pool: DummyPool | ThreadPool = DummyPool() if cpus == 1 else ThreadPool(cpus)
    with pool:
        args = [(path, warn, with_annotations) for path in template_paths]
        try:
            template_iterator = pool.starmap(_load_and_annotate, args)
            for loaded_template in template_iterator:
                yield loaded_template
        except ValueError as exc:
            raise exc


# TODO adapt this too for AnnotatedTemplate objects
def check_template(template: Template, warn: bool = True) -> bool:
    # TODO improve this and write tests
    if warn:
        checker = True

        # Raise warnings if some properties could not be annotated!
        if not template.ec:
            checker = False
            warnings.warn("Could not find EC number annotations")

        if not template.cath:
            checker = False
            warnings.warn("Could not find CATH annotations")

        if template.pdb_id:
            # check overlap between sifts mapping and CATH, EC annotations
            # Source: pdb to sifts mapping which maps CATH to pdb chain IDs and UniProt IDs, sifts also provides UniProt to EC mapping
            # Since Template may contain residues from multiple chains
            pdbchains = set()
            for res in template.residues:
                pdbchains.add(template.pdb_id + res.chain_id)

            # Iterating of all pdbchains which were part of the Template
            for pdbchain in pdbchains:
                # also include EC annotations from PDB-SIFTS
                subdict = template._PDB_SIFTS.get(pdbchain, None)
                if subdict:
                    # if template.cath and subdict['cath']:
                    #     if not set(template.cath) & set(subdict['cath']):
                    #         warnings.warn(f'Did not find an intersection of CATH domains as annotated by the M-CSA ID {template.mcsa_id} with {template.cath} versus PDF-SIFTS {template._PDB_SIFTS[pdbchain]['cath']} for template {filepath} from PDB ID {template.pdb_id}')
                    sifts_uniprot = subdict["uniprot_id"]  # type: ignore
                    if template.uniprot_id != sifts_uniprot:
                        checker = False
                        warnings.warn(
                            f"Different UniProt Accessions {template.uniprot_id} and {sifts_uniprot} found"
                        )

        return checker
    else:
        return True



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from collections import OrderedDict
import json
from typing import Optional, List, Dict, Tuple

# NOTE I know these could be frozen dataclasses too
# but that makes it annoying to apply manually fixes post init
# which is unfortunately necessary sometimes


@dataclass(frozen=False)
class HomologousResidue:
    """
    Base class for a PDB residue identical or homologous to a template residue.

    Attributes:
        code : `str` Three letter amino-acid code
        resid : `int` residue id in the PDB
        auth_resid : `int` Author assigned residue id

    """

    code: str  # three letter code
    resid: int
    auth_resid: int
    # domain_name: Optional[str]
    # domain_cath_id: Optional[str]

    def __post_init__(self):
        object.__setattr__(self, "code", self.code.upper())


@dataclass(frozen=False)
class ReferenceCatalyticResidue(HomologousResidue):
    """
    Class for a reference PDB residue for a template residue.

    Attributes:
        function_location_abv : `str` Functional part of the residue. Empty string means side chain. Else "main", "main-C", "main-N" or "ptm"
        ptm : `str` Empty string or for post translationally modified residues, the 3-letter code of the ptm in the PDB
        roles : `List[str]` List of EMO codes describing functional roles
        roles_summary : `List[str]` List text discriptions of functional roles

    """

    function_location_abv: Optional[str]
    ptm: Optional[str]
    roles: List[str]
    roles_summary: List[str]


@dataclass(frozen=False)
class NonReferenceCatalyticResidue(HomologousResidue):
    """
    Class for a non-reference PDB residue for a template residue.

    Attributes:
        reference : `Tuple[int, str, int]` : Tuple of mcsa_id, pdb_id, residue_index to find the reference

    """

    reference: Tuple[int, str, int]  # can link to with mcsa_id, pdb_id, residue index


@dataclass(frozen=False)
class HomologousPDB:
    """
    Class for a PDB Entry homologous to or identical to the structure from which a template was generated.

    Attributes:
        mcsa_id : `int` : M-CSA ID
        reference_pdbchain : `str` : PDB identifier of the M-CSA reference PDB structure
        is_reference : `bool` : If this structure is an M-CSA reference
        pdb_id : `str` : PDB identifier of the PDB structure
        chain_name : `str` : Chain identifier in the PDB
        assembly_chain_name : `str` : Author assigned assembly chain identifier
        assembly : `int` : Assembly number (usually but not always the biological assembly)
        residues : `Dict[int, HomologousResidue]` : Dictionary mapping residue indices to residues. If "is_reference" then values are `ReferenceCatalyticResidue` else they are `NonReferenceCatalyticResidue`

    """

    # because we iteratively add residues, the HomologousPDB dataclass
    # cannot be frozen
    mcsa_id: int
    reference_pdbchain: str
    is_reference: bool
    pdb_id: str
    chain_name: str
    assembly_chain_name: str
    assembly: int
    residues: Dict[int, HomologousResidue]  # can fully contain

    def __post_init__(self):
        object.__setattr__(self, "pdb_id", self.pdb_id.lower())

    def _state(self) -> Tuple:
        return (
            self.mcsa_id,
            self.reference_pdbchain,
            self.is_reference,
            self.pdb_id,
            self.chain_name,
            self.assembly_chain_name,
            self.assembly,
            tuple(self.residues),
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, HomologousPDB):
            self_state = self._state()
            other_state = other._state()
            return self_state == other_state
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, HomologousPDB):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self) -> int:
        return hash((type(self), self._state()))


# @dataclass(frozen=True)
# class Entry:
#     uniprot: str
#     reference_pdb: HomologousPDB
#     number_of_residues: int


def load_mcsa_catalytic_residue_homologs_info(
    outdir: Path,
) -> Dict[str, Dict[str, HomologousPDB]]:

    file = outdir.joinpath("catalytic_residue_homologs_information.json")
    # file is probably in Path(outdir, "catalytic_residue_homologs_information.json")
    with open(file, "r") as f:
        data = json.load(f)

    out: OrderedDict = OrderedDict()
    for mcsa_id, pdb_entries in data.items():
        mcsa_id = int(mcsa_id)
        out[mcsa_id] = OrderedDict()
        for pdb_id, pdb_data in pdb_entries.items():
            residues: Dict[int, HomologousResidue] = {}
            # If it is a reference structure,
            # for multimetic catalytic sites, references may be mixed
            if pdb_data["is_reference"]:
                for index, residue in pdb_data["residues"].items():
                    index = int(index)
                    try:
                        residues[index] = ReferenceCatalyticResidue(
                            code=residue["code"],
                            resid=residue["resid"],
                            auth_resid=residue["auth_resid"],
                            function_location_abv=residue["function_location_abv"],
                            ptm=residue["ptm"],
                            roles=residue["roles"],
                            roles_summary=residue["roles_summary"],
                        )
                    except KeyError:
                        residues[index] = NonReferenceCatalyticResidue(
                            code=residue["code"],
                            resid=residue["resid"],
                            auth_resid=residue["auth_resid"],
                            reference=tuple(residue["reference"]),
                        )
            # If it is not a reference, all residues are NonReference
            else:
                for index, residue in pdb_data["residues"].items():
                    index = int(index)
                    residues[index] = NonReferenceCatalyticResidue(
                        code=residue["code"],
                        resid=residue["resid"],
                        auth_resid=residue["auth_resid"],
                        reference=tuple(residue["reference"]),
                    )
            out[mcsa_id][pdb_id] = HomologousPDB(
                mcsa_id=mcsa_id,
                reference_pdbchain=pdb_data["reference_pdbchain"],
                is_reference=pdb_data["is_reference"],
                pdb_id=pdb_data["pdb_id"],
                chain_name=pdb_data["chain_name"],
                assembly_chain_name=pdb_data["assembly_chain_name"],
                assembly=pdb_data["assembly"],
                residues=residues,
            )

    return out



from __future__ import annotations

import argparse
from pathlib import Path
from typing import Optional, List
import warnings
import sys
import os

import rich.console

from enzymm import __version__
from enzymm.template import load_templates
from enzymm.jess_run import Matcher, Match, load_molecules


def build_parser() -> argparse.ArgumentParser:
    """Parse Arguments with Argparse. Returns args object"""

    class ReadListAction(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            with values.open("r") as f:
                for line in f:
                    dest = getattr(namespace, self.dest)
                    dest.append(Path(line.strip()))

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        usage="""
            Minimal use: enzymm -i query.pdb -o result.tsv
            
            Recommended use: enzymm -l query_pdbs.list -o results.tsv -v --pdbs --include-query -n -1
            """,
        description="EnzyMM - EnzymeMotifMiner - version {__version__}\nGeometric matching of catalytic motifs in protein structures.",
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"enzymm {__version__}"
    )
    parser.add_argument(
        "-o",
        "--output",
        required=True,
        type=Path,
        help="Output tsv file to which results should get written",
    )
    parser.add_argument(
        "--pdbs",
        type=Path,
        help="Output directory to which results should get written",
        default=None,
    )

    # inputs: either a list of paths, or directly a path (or any combination)
    parser.add_argument(
        "-i",
        "--input",
        type=Path,
        help="File path to a PDB file to use as query",
        action="append",
        dest="files",
        default=[],
    )
    parser.add_argument(
        "-l",
        "--list",
        type=Path,
        help="File containing a list of PDB files to read",
        action=ReadListAction,
        dest="files",
    )

    # optional arguments
    parser.add_argument(
        "-j",
        "--jess",
        nargs=3,
        default=None,
        type=float,
        help="Fixed Jess parameters for all templates. Jess space seperated parameters rmsd, distance, max_dynamic_distance",
    )
    parser.add_argument(
        "-t",
        "--template-dir",
        type=Path,
        default=None,
        help="Path to directory containing jess templates. This directory will be recursively searched.",
    )
    parser.add_argument(
        "-c",
        "--conservation-cutoff",
        type=float,
        default=0,
        help="Atoms with a value in the B-factor column below this cutoff will be excluded form matching to the templates. Useful for predicted structures.",
    )
    parser.add_argument(
        "-n",
        "--n-jobs",
        type=int,
        default=(
            len(os.sched_getaffinity(0))
            if sys.platform == "linux"
            else os.cpu_count() or 1
        ),  # len(os.sched_getaffinity(0)) doesnt work on mac/win
        help="The number of threads to run in parallel. Pass 0 to select all. Negative numbers: leave this many threads free.",
    )

    # Boolean optional arguments
    parser.add_argument(
        "-v",
        "--verbose",
        default=False,
        action="store_true",
        help="If process information and time progress should be printed to the command line",
    )
    parser.add_argument(
        "-w",
        "--warn",
        default=False,
        action="store_true",
        help="If warings about bad template processing or suspicous and missing annotations should be raised",
    )
    parser.add_argument(
        "-q",
        "--include-query",
        default=False,
        action="store_true",
        help="Include the query structure together with the hits in the pdb output",
    )
    parser.add_argument(
        "-u",
        "--unfiltered",
        default=False,
        action="store_true",
        help="If set, matches which logistic regression predicts as false based on RMSD and resdiue orientation will be retained. By default, matches predicted as false are removed.",
    )
    parser.add_argument(
        "--transform",
        default=False,
        action="store_true",
        help="If set, will transform the coordinate system of the hits to that of the template in the pdb output",
    )
    parser.add_argument(
        "--skip-smaller-hits",
        default=False,
        action="store_true",
        help="If set, will not search with smaller templates if larger templates have already found hits.",
    )
    parser.add_argument(
        "--match-small-templates",
        default=False,
        action="store_true",
        help="If set, templates with less then 3 defined sidechain residues will still be matched.",
    )
    parser.add_argument(
        "--skip-annotation",
        default=False,
        action="store_true",
        help="If set, M-CSA derived templates will NOT be annotated with extra information.",
    )
    return parser


def main(argv: Optional[List[str]] = None, stderr=sys.stderr):

    parser = build_parser()
    args = parser.parse_args(args=argv)
    if not args.files:
        raise ValueError("No input files were passed. Use -i and/or -l.")

    jess_params = None
    if args.jess:
        jess_params_list = [i for i in args.jess]
        # jess parameters
        # we use different parameters for different template residue numbers - higher number more generous parameters
        rmsd = jess_params_list[0]  # in Angstrom, typcically set to 2
        distance = jess_params_list[
            1
        ]  # in Angstrom between 1.0 and 1.5 - lower is more strict. This changes with template size
        max_dynamic_distance = jess_params_list[
            2
        ]  # if equal to distance dynamic is off: this option is currenlty dysfunctional

        known_distances = [
            float(i) for i in Match._logistic_regression_models["3"].keys()
        ]
        if distance not in known_distances and not args.unfiltered:
            raise ValueError(
                "Filtering paraterms only established for pairwise distances in 0.7, 0.8, ... , 2.0A "
            )

        jess_params = {
            3: {
                "rmsd": rmsd,
                "distance": distance,
                "max_dynamic_distance": max_dynamic_distance,
            },
            4: {
                "rmsd": rmsd,
                "distance": distance,
                "max_dynamic_distance": max_dynamic_distance,
            },
            5: {
                "rmsd": rmsd,
                "distance": distance,
                "max_dynamic_distance": max_dynamic_distance,
            },
            6: {
                "rmsd": rmsd,
                "distance": distance,
                "max_dynamic_distance": max_dynamic_distance,
            },
            7: {
                "rmsd": rmsd,
                "distance": distance,
                "max_dynamic_distance": max_dynamic_distance,
            },
            8: {
                "rmsd": rmsd,
                "distance": distance,
                "max_dynamic_distance": max_dynamic_distance,
            },
        }

    try:
        molecules = load_molecules(
            molecule_paths=args.files,
            conservation_cutoff=args.conservation_cutoff,
            warn=args.warn,
        )

        templates = list(
            load_templates(
                template_dir=args.template_dir,
                warn=args.warn,
                verbose=args.verbose,
                cpus=args.n_jobs,
                with_annotations=not args.skip_annotation,
            )
        )

        ############ Initialize Matcher object ################################
        matcher = Matcher(
            templates=templates,
            jess_params=jess_params,
            conservation_cutoff=args.conservation_cutoff,
            warn=args.warn,
            verbose=args.verbose,
            skip_smaller_hits=args.skip_smaller_hits,
            match_small_templates=args.match_small_templates,
            cpus=args.n_jobs,
            filter_matches=not args.unfiltered,
            console=rich.console.Console(file=stderr),
        )

        ############ Call Matcher.run ##########################################
        processed_molecules = matcher.run(molecules=molecules)

        ######### Writing Output ##########################################
        out_tsv = args.output

        if not out_tsv.parent.exists():
            out_tsv.parent.mkdir(parents=True, exist_ok=True)
            if args.warn:
                warnings.warn(
                    f"{out_tsv.parent.resolve()} directory tree to output tsv file did not exist and was created"
                )
        elif out_tsv.exists() and args.warn:
            warnings.warn(
                f"The specified output tsv file {out_tsv.resolve()} already exists and will be overwritten!"
            )

        if args.verbose:
            print(f"Writing output to {out_tsv.resolve()}")
            print(
                f"Matches predicted by logistic regression as false are {'' if args.unfiltered else 'not '}reported"
            )

        with open(out_tsv, "w", newline="", encoding="utf-8") as tsvfile:
            tsvfile.write(f"# Version {__version__}\n")
            for index, (molecule, matches) in enumerate(processed_molecules.items()):
                for jndex, match in enumerate(matches):
                    i = index + jndex
                    match.index = jndex + 1  # 1 indexed matches per query
                    match.dump(
                        tsvfile,
                        header=(i == 0),
                        predict_correctness=not args.unfiltered,
                    )  # one line per match, write header only for the first match

        # TODO with multiple matches per query we have two options:
        # with --include-query and not --transform we can write them all into one pdb
        # in this case, we dont need to do any furhter processing.

        # with --transform this doesnt make much sense.
        # here we would like to write each match in its own file
        # to a seperate folder named after the query
        # then we would include also the template and
        # the query and do all the alignment processing

        # TODO print warning saying that setting the args.transform
        # option doesnt make sense when
        # searching with multiple templates in the current setup.
        # as described above, we would want seperate output files per match

        def write_hits2pdb(matches: List[Match], filename: str, outdir: Path):
            # make sure molecule().id is unique!
            with open(
                Path(outdir, f"{filename}_matches.pdb"), "w", encoding="utf-8"
            ) as pdbfile:
                if (
                    args.include_query
                ):  # write the molecule structure to the top of the pdb output too
                    for i, match in enumerate(matches):
                        match.dump2pdb(
                            pdbfile, include_query=(i == 0), transform=args.transform
                        )
                else:
                    for match in matches:
                        match.dump2pdb(pdbfile, transform=args.transform)

        if args.pdbs:
            args.pdbs.mkdir(parents=True, exist_ok=False)
            for molecule, matches in processed_molecules.items():
                write_hits2pdb(matches=matches, filename=molecule.id, outdir=args.pdbs)  # type: ignore

    except IsADirectoryError as exc:
        print("File is a directory:", exc.filename, file=stderr)
        return exc.errno

    except FileNotFoundError as exc:
        print("Failed to find file:", exc.filename, file=stderr)
        return exc.errno

    except FileExistsError as exc:
        print("File already exists:", exc.filename, file=stderr)
        return exc.errno

    return 0


if __name__ == "__main__":
    sys.exit(main())
